# The objective of this class is to create a dictionary of video streams that can
# be handed off to a synchronizer, which will then interact with the streams
# as though they were live video
# this is useful for two purposes:
#   1: future testing (don't have to keep recording live video)
#   2: future off-line processing of pre-recorded video.

import pyxy3d.logger
import logging

logger = pyxy3d.logger.get(__name__)
logger.setLevel(logging.INFO)

from pathlib import Path
from queue import Queue
from threading import Thread, Event
import toml

import cv2
from time import perf_counter, sleep
import pandas as pd
import numpy as np

from pyxy3d.interface import FramePacket, Tracker, Stream
from pyxy3d.trackers.tracker_enum import TrackerEnum
from pyxy3d.cameras.live_stream import Stream
from pyxy3d.cameras.camera_array import CameraData
from pyxy3d.configurator import Configurator


class RecordedStream(Stream):
    """
    Analogous to the live stream, this will place frames on a queue
    These can then be harvested and synchronized by a Synchronizer
    Within the stream, point detection occurs.
    """

    def __init__(
        self, camera: CameraData, directory: Path, fps_target:int=6, tracker: Tracker = None
    ):
        # self.port = port
        self.directory = directory
        self.camera = camera
        self.port = camera.port

        if tracker is not None:
            self.tracker = tracker
            self.track_points = True
        else:
            self.track_points = False

        video_path = str(Path(self.directory, f"port_{self.port}.mp4"))
        self.capture = cv2.VideoCapture(video_path)

        self.stop_event = Event()

        self.subscribers = []

        synched_frames_history_path = str(
            Path(self.directory, f"frame_time_history.csv")
        )
        synched_frames_history = pd.read_csv(synched_frames_history_path)

        self.port_history = synched_frames_history[
            synched_frames_history["port"] == self.port
        ]
        self.start_frame_index = self.port_history["frame_index"].min()
        self.last_frame_index = self.port_history["frame_index"].max()

        # initializing to something to avoid errors elsewhere
        self.frame_index = 0
        self.frame_time = 0
        self.set_fps_target(fps_target)

    def set_tracking_on(self, track: bool):
        if track:
            logger.info(f"Turning tracking on for recorded stream {self.port}")
            self.track_points.set()
        else:
            logger.info(f"Turning tracking off for recorded stream {self.port}")
            self.track_points.clear()

    def subscribe(self, queue: Queue):
        if queue not in self.subscribers:
            logger.info(f"Adding queue to subscribers at recorded stream {self.port}")
            self.subscribers.append(queue)
            logger.info(f"...now {len(self.subscribers)} subscriber(s) at {self.port}")
        else:
            logger.warn(
                f"Attempted to subscribe to recorded stream at port {self.port} twice"
            )

    def unsubscribe(self, queue: Queue):
        if queue in self.subscribers:
            logger.info(
                f"Removing subscriber from queue at recorded stream {self.port}"
            )
            self.subscribers.remove(queue)
            logger.info(
                f"{len(self.subscribers)} subscriber(s) remain at recorded stream {self.port}"
            )
        else:
            logger.warn(
                f"Attempted to unsubscribe to recorded stream that was not subscribed to\
                at port {self.port} twice"
            )

    def set_fps_target(self, fps):
        self.fps = fps
        if self.fps is None:
            self.milestones = None
        else:
            milestones = []
            for i in range(0, fps):
                milestones.append(i / fps)
            logger.info(f"Setting fps to {self.fps}")
            self.milestones = np.array(milestones)

    def wait_to_next_frame(self):
        """
        based on the next milestone time, return the time needed to sleep so that
        a frame read immediately after would occur when needed
        """

        time = perf_counter()
        fractional_time = time % 1
        all_wait_times = self.milestones - fractional_time
        future_wait_times = all_wait_times[all_wait_times > 0]

        if len(future_wait_times) == 0:
            return 1 - fractional_time
        else:
            return future_wait_times[0]

    def play_video(self):
        self.thread = Thread(target=self.process_frames, args=[], daemon=True)
        self.thread.start()

    def process_frames(self):
        """
        Places FramePacket on the out_q, mimicking the behaviour of the LiveStream.
        """

        self.frame_index = self.start_frame_index
        logger.info(f"Beginning playback of video for port {self.port}")
        while not self.stop_event.is_set():
            current_frame = self.port_history["frame_index"] == self.frame_index
            self.frame_time = self.port_history[current_frame]["frame_time"]
            self.frame_time = float(self.frame_time)

            spinlock_looped = False
            while len(self.subscribers) == 0 and not self.stop_event.is_set():
                if not spinlock_looped:
                    logger.info(f"Spinlock initiated at port {self.port}")
                    spinlock_looped = True
                sleep(0.5)
            if spinlock_looped == True:
                logger.info(f"Spinlock released at port {self.port}")

            if self.milestones is not None:
                sleep(self.wait_to_next_frame())

            success, self.frame = self.capture.read()

            if not success:
                break

            if self.track_points:
                self.point_data = self.tracker.get_points(self.frame, self.port, self.camera.rotation_count)
            else:
                self.point_data = None

            frame_packet = FramePacket(
                port=self.port,
                frame_time=self.frame_time,
                frame=self.frame,
                frame_index=self.frame_index,
                points=self.point_data,
                draw_instructions=self.tracker.draw_instructions,
            )
            
            
            # cv2.imshow(str(self.port), frame_packet.frame_with_points)
            # key = cv2.waitKey(1)
            # if key == ord("q"):
            #     cv2.destroyAllWindows()
            #     break

            logger.debug(
                f"Placing frame on reel {self.port} for frame time: {self.frame_time} and frame index: {self.frame_index}"
            )

            for q in self.subscribers:
                q.put(frame_packet)

            # self.out_q.put(frame_packet)
            self.frame_index += 1

            if self.frame_index >= self.last_frame_index:
                logger.info(f"Ending recorded playback at port {self.port}")
                # time of -1 indicates end of stream
                frame_packet = FramePacket(self.port, -1, None, None)

                for q in self.subscribers:
                    q.put(frame_packet)
                break


class RecordedStreamPool:
    def __init__(
        self,
        directory: Path,
        config: Configurator,
        fps_target=6,
        tracker: Tracker = None,
    ):
        self.streams = {}
        self.camera_array = config.get_camera_array()

        for port, camera in self.camera_array.cameras.items():
            # tracker: Tracker = tracker.value()
            self.streams[port] = RecordedStream(
                camera, directory, fps_target=fps_target, tracker=tracker
            )

    def play_videos(self):
        for port, stream in self.streams.items():
            stream.play_video()


def get_configured_camera_data(config_path, intrinsics_only=True):
    """
    return a list of CameraData objects that is built from the config
    file that is found in the directory. This will be the same
    file where the mp4 files are located.
    """

    with open(config_path, "r") as f:
        config = toml.load(config_path)

    all_camera_data = {}
    for key, params in config.items():
        if key.startswith("cam"):
            if not params["ignore"]:
                port = params["port"]

                if intrinsics_only:
                    logger.info(
                        f"Adding intrinisic camera {port} to calibrated camera array..."
                    )

                    cam_data = CameraData(
                        port=port,
                        size=params["size"],
                        rotation_count=params["rotation_count"],
                        error=params["error"],
                        matrix=np.array(params["matrix"]),
                        distortions=np.array(params["distortions"]),
                        exposure=params["exposure"],
                        grid_count=params["grid_count"],
                        ignore=params["ignore"],
                        verified_resolutions=params["verified_resolutions"],
                    )
                else:
                    logger.info(
                        f"Adding intrinsic and extrinsic camera {port} to calibrated camera array..."
                    )
                    cam_data = CameraData(
                        port=port,
                        size=params["size"],
                        rotation_count=params["rotation_count"],
                        error=params["error"],
                        matrix=np.array(params["matrix"]),
                        distortions=np.array(params["distortions"]),
                        exposure=params["exposure"],
                        grid_count=params["grid_count"],
                        ignore=params["ignore"],
                        verified_resolutions=params["verified_resolutions"],
                        translation=np.array(params["translation"]),
                        rotation=np.array(params["rotation"]),
                    )

                all_camera_data[port] = cam_data

    return all_camera_data


if __name__ == "__main__":
    from pyxy3d.trackers.charuco_tracker import CharucoTracker
    from pyxy3d.cameras.synchronizer import Synchronizer
    from pyxy3d.calibration.charuco import Charuco

    from pyxy3d import __root__

    recording_directory = Path(__root__, "tests", "sessions", "post_monocal")

    charuco = Charuco(
        4, 5, 11, 8.5, aruco_scale=0.75, square_size_overide_cm=5.25, inverted=True
    )

    tracker = CharucoTracker(charuco)

    cameras = get_configured_camera_data(recording_directory)

    recorded_stream_pool = RecordedStreamPool(
        recording_directory, tracker_factory=tracker
    )
    syncr = Synchronizer(recorded_stream_pool.streams, fps_target=None)
    recorded_stream_pool.play_videos()

    syncr.subscribe_to_streams()

    in_q = Queue(-1)
    syncr.subscribe_to_sync_packets(in_q)

    while not syncr.frames_complete:
        sleep(0.03)
        sync_packet = in_q.get()
        for port, frame_packet in sync_packet.frame_packets.items():
            if frame_packet:
                if frame_packet.frame_time == -1:
                    break  # end of frames
                cv2.imshow(f"Port {port}", frame_packet.frame_with_points)

        key = cv2.waitKey(1)

        if key == ord("q"):
            # cv2.destroyAllWindows()
            break

    cv2.destroyAllWindows()



# %%

import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

from pathlib import Path
import numpy as np
from dataclasses import dataclass
import cv2
from enum import Enum, auto

CAMERA_PARAM_COUNT = 6

@dataclass
class CameraData:
    """
    A place to hold the calibration data associated with a camera that has been populated from a config file.
    For use with final setting of the array and triangulation, but no actual camera management.
    Not loving the way this is implemented as an adjunct to the Camera object, but here we are
    """

    port: int
    size: tuple
    rotation_count: int
    error: float  # the RMSE of reprojection associated with the intrinsic calibration
    matrix: np.ndarray
    distortions: np.ndarray  #
    exposure: int
    grid_count: int
    ignore: bool
    verified_resolutions: np.ndarray
    translation: np.ndarray =None # camera relative to world
    rotation: np.ndarray = None# camera relative to world

    @property
    def transformation(self):
        """"
        Rotation and transformation combined to allow 
        """
        
        t = np.hstack([self.rotation, np.expand_dims(self.translation, 1)])
        t = np.vstack([t, np.array([0,0,0,1], np.float32)])
        return t 
   
    @transformation.setter 
    def transformation(self, t: np.ndarray):
        self.rotation = t[0:3,0:3]
        self.translation = t[0:3,3]
        logger.info(f"Rotation and Translation being updated to {self.rotation} and {self.translation}")
    
    @property  
    def projection_matrix(self):
        return self.matrix @ self.transformation[0:3,:]
         
    def extrinsics_to_vector(self):
        """
        Converts camera parameters to a numpy vector for use with bundle adjustment.
        """
        # rotation of the camera relative to the world
        rotation_rodrigues = cv2.Rodrigues(self.rotation)[0]  # elements 0,1,2
        port_param = np.hstack([rotation_rodrigues[:, 0], self.translation])

        return port_param

    def extrinsics_from_vector(self, row):
        """
        Takes a vector in the same format that is output of `extrinsics_to_vector` and updates the camera
        """

        # convert back to world frame of reference
        self.rotation = cv2.Rodrigues(row[0:3])[0]
        self.translation = np.array([row[3:6]], dtype=np.float64)[0]


@dataclass
class CameraArray:
    """The plan is that this will expand to become an interface for setting the origin.
    At the moment all it is doing is holding a dictionary of CameraData objects"""

    cameras: dict

    # least_sq_result = None # this may be a vestige of an old way of doing things. remove if commenting out doesn't break anything.

    @property
    def port_index(self):
        """
        Provides a dictionary mapping the camera port to an index. Generally,this
        will match the camera ports 1:1, but will be different when a camera 
        is being ignored. Used to manage reference to camera parameters in xy_reprojection_error
        used within the least_squares optimization of the capture volume. 
        """
        not_ignored_ports = [port for port, cam in self.cameras.items() if not cam.ignore]
        not_ignored_ports.sort()
        not_ignored_indices = [i for i in range(len(not_ignored_ports))]
        port_indices = {port:i for port, i in zip(not_ignored_ports,not_ignored_indices)}                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                       
        
        return port_indices
   
    @property 
    def index_port(self):
        return {value: key for key, value in self.port_index.items()}
    
    def get_extrinsic_params(self):
        """for each camera build the CAMERA_PARAM_COUNT element parameter index
        camera_params with shape (n_cameras, CAMERA_PARAM_COUNT)
        contains initial estimates of parameters for all cameras.
        First 3 components in each row form a rotation vector (https://en.wikipedia.org/wiki/Rodrigues%27_rotation_formula),
        next 3 components form a translation vector
        """

        camera_params = None
        # ensure that parameters are built up in order of the corresponding index
        for index in sorted(self.index_port.keys()):
            port = self.index_port[index]
            cam = self.cameras[port]
            port_param = cam.extrinsics_to_vector()

            if camera_params is None:
                camera_params = port_param
            else:
                camera_params = np.vstack([camera_params, port_param])

        return camera_params

    def update_extrinsic_params(self, least_sq_result_x:np.array):

        n_cameras = len(self.port_index)
        n_cam_param = 6  # 6 DoF
        flat_camera_params = least_sq_result_x[0 : n_cameras * n_cam_param]
        new_camera_params = flat_camera_params.reshape(n_cameras, n_cam_param)

        # update camera array with new positional data
        for index in range(len(new_camera_params)):
            port = self.index_port[index]  # correct in case ignoring a camera
            cam_vec = new_camera_params[index, :]
            self.cameras[port].extrinsics_from_vector(cam_vec)
        
    @property
    def extrinsics_calibrated(self)->bool:
       
        # assume extrinsics calibrated and provide otherwise
        full_extrinsics = True 
        for port, cam in self.cameras.items():
            if cam.rotation is None or cam.translation is None:
                full_extrinsics = False

        return full_extrinsics
        
            
            

class CalibrationStage(Enum):
    """
    NOTE: this is not currently implemented. I was planning to set this up
    in dev_post_process_eligible_check as a way to help manage the flow of the
    GUI (it would be important to now the status of teh calibration in an easy way)
    Unfortunately this begins to touch up on how/when the session loads the camera array,
    and that's something that I'm not sure is fully relevant right now.
    The camera array is not initialized until it is estimated via bundle adjustment.
    Prior to that it is only a dictionary of cameras. 
    
    Consider removing...

    """
    NO_INTRINSICS = auto()
    PARTIAL_INTRINSICS = auto()
    FULL_INTRINSICS = auto()
    EXTRINSICS = auto()


# the PointHistory object requires a highly particular format
# all of the data is ultimately embedded in the initial camera array configuration
# and the calibration point data. These functions transform those two 
# things into a PointHistory object that can be used to optimize the CaptureVolume
import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

from itertools import combinations
import pandas as pd
import numpy as np
from pathlib import Path
import time

from pyxy3d import __root__
from pyxy3d.interface import PointPacket, FramePacket, SyncPacket
from pyxy3d.cameras.camera_array import CameraArray
from pyxy3d.triangulate.array_stereo_triangulator import ArrayStereoTriangulator

from pyxy3d.triangulate.stereo_points_builder import (
    StereoPointsBuilder,
    StereoPointsPacket,
    SynchedStereoPointsPacket,
)



def get_stereotriangulated_table(camera_array:CameraArray, point_data_path:Path) -> pd.DataFrame:

    logger.info(f"Beginning to create stereotriangulated points from data stored at {point_data_path}")
    point_data = pd.read_csv(point_data_path)

    xy_sync_indices = point_data["sync_index"].to_numpy()
    sync_indices = np.unique(xy_sync_indices)


    xy_camera_indices = point_data["port"].to_numpy()
    # ports = np.unique(xy_camera_indices)
    ports = [key for key in camera_array.port_index.keys()]

    paired_point_builder = StereoPointsBuilder(ports)

    # Create the infrastructure for the pairwise triangulation
    array_triangulator = ArrayStereoTriangulator(camera_array)
    stereotriangulated_table = None

    logger.info(f"Begin reconstructing SyncPackets and SynchedStereoPairs... ")
    for sync_index in sync_indices:
        if sync_index%25==0:
            logger.info(f"Processing stereotriangulation estimates...currently at sync index {sync_index}")
        # pull in the data that shares the same sync index
        port_points = point_data.query(f"sync_index == {sync_index}")

        # initialize a dict to hold all the frame packets
        frame_packets = {}

        for port in ports:
            # Create the Frame packet for each port at this sync_index
            # and roll up into dictionary
            if port in port_points["port"].unique():
                points = port_points.query(f"port == {port}")
                frame_time = points["frame_time"].iloc[0]
                frame_index = points["frame_index"].iloc[0]

                point_id = points["point_id"].to_numpy()

                img_loc_x = points["img_loc_x"].to_numpy()
                img_loc_y = points["img_loc_y"].to_numpy()
                img_loc = np.vstack([img_loc_x, img_loc_y]).T

                obj_loc_x = points["obj_loc_x"].to_numpy()
                obj_loc_y = points["obj_loc_y"].to_numpy()
                obj_loc = np.vstack([obj_loc_x, obj_loc_y]).T

                point_packet = PointPacket(point_id, img_loc, obj_loc)
                frame_packet = FramePacket(
                    port, frame_time, None, frame_index, point_packet
                )
                frame_packets[port] = frame_packet
            else:
                frame_packets[port] = None

        # create the sync packet for this sync index
        sync_packet = SyncPacket(sync_index, frame_packets)

        # get the paired point packets for all port pairs at this sync index
        synched_stereo_points: SynchedStereoPointsPacket = (
            paired_point_builder.get_synched_paired_points(sync_packet)
        )
        # print(synched_paired_points)
        array_triangulator.triangulate_synched_points(synched_stereo_points)

        for pair in synched_stereo_points.pairs:
            triangulated_pair: StereoPointsPacket = (
                synched_stereo_points.stereo_points_packets[pair]
            )
            if triangulated_pair is not None:
                if stereotriangulated_table is None:
                    stereotriangulated_table = triangulated_pair.to_table()
                else:
                    new_table = triangulated_pair.to_table()
                    for key, value in new_table.items():
                        stereotriangulated_table[key].extend(value)

    logger.info(f"Saving stereotriangulated_points.csv to {point_data_path.parent} for inspection")
    stereotriangulated_table = pd.DataFrame(stereotriangulated_table)
    stereotriangulated_table.to_csv(
        Path(point_data_path.parent, "stereotriangulated_points.csv")
    )

    logger.info("Returning dataframe of stereotriangulated points to caller")
    
    return stereotriangulated_table


import logging
import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)
if __name__ == "__main__":
    logger.setLevel(logging.DEBUG)

from queue import Queue
from threading import Thread, Event
import cv2
import sys
from pathlib import Path
import pandas as pd
import numpy as np
from dataclasses import dataclass
from itertools import combinations
from pyxy3d.cameras.synchronizer import Synchronizer
from pyxy3d.trackers.charuco_tracker import CharucoTracker
from pyxy3d.interface import SyncPacket


class StereoPointsBuilder:
    def __init__(self, ports: list):

        self.ports = ports
        self.pairs = [(i, j) for i, j in combinations(self.ports, 2) if i < j]

    def get_stereo_points_packet(self, sync_index, port_A, points_A, port_B, points_B):

        # get ids in common
        if len(points_A.point_id) > 0 and len(points_B.point_id) > 0:
            common_ids = np.intersect1d(points_A.point_id, points_B.point_id)
        else:
            common_ids = np.array([])

        if len(common_ids) == 0:
            packet = None
        else:
            # common_ids = common_ids[:,0]
            # for both ports, get the indices of the common ids
            sorter_A = np.argsort(points_A.point_id)
            shared_indices_A = sorter_A[
                np.searchsorted(points_A.point_id, common_ids, sorter=sorter_A)
            ]
            shared_indices_A

            sorter_B = np.argsort(points_B.point_id)
            shared_indices_B = sorter_B[
                np.searchsorted(points_B.point_id, common_ids, sorter=sorter_B)
            ]
            shared_indices_B

            packet = StereoPointsPacket(
                sync_index=sync_index,
                port_A=port_A,
                port_B=port_B,
                common_ids=common_ids,
                img_loc_A=points_A.img_loc[shared_indices_A],
                img_loc_B=points_B.img_loc[shared_indices_B],
            )

            logger.debug(
                f"Points in common for ports ({port_A}, {port_B}): {common_ids}"
            )

        return packet

    def get_synched_paired_points(self, sync_packet: SyncPacket):

        # will be populated with dataframes of:
        # id | img_x | img_y | board_x | board_y
        sync_index = sync_packet.sync_index
        paired_points_packets = {}

        for pair in self.pairs:
            port_A = pair[0]
            port_B = pair[1]

            if (
                sync_packet.frame_packets[port_A] is not None
                and sync_packet.frame_packets[port_B] is not None
            ):

                points_A = sync_packet.frame_packets[port_A].points
                points_B = sync_packet.frame_packets[port_B].points

                paired_points: StereoPointsPacket = self.get_stereo_points_packet(
                    sync_index, port_A, points_A, port_B, points_B
                )

                paired_points_packets[pair] = paired_points

        return SynchedStereoPointsPacket(sync_index, paired_points_packets)


@dataclass
class StereoPointsPacket:
    """The points shared by two FramePointsPackets"""

    sync_index: int

    port_A: int
    port_B: int

    common_ids: np.ndarray
    img_loc_A: np.ndarray
    img_loc_B: np.ndarray

    # a place to hold the pairwise triangulated value down the line
    xyz: np.ndarray = None

    @property
    def pair(self):
        return (self.port_A, self.port_B)

    def to_table(self):
        # table will be in the form of a dictionary of lists of equal length
        table = {}

        point_count = len(self.common_ids)

        table["pair"] = [self.pair] * point_count
        table["port_A"] = [self.port_A] * point_count
        table["port_B"] = [self.port_B] * point_count
        table["sync_index"] = [self.sync_index] * point_count
        table["point_id"] = list(self.common_ids)
        table["x_pos"] = list(self.xyz[:, 0])
        table["y_pos"] = list(self.xyz[:, 1])
        table["z_pos"] = list(self.xyz[:, 2])
        table["x_A"] = list(self.img_loc_A[:, 0])
        table["y_A"] = list(self.img_loc_A[:, 1])
        table["x_B"] = list(self.img_loc_B[:, 0])
        table["y_B"] = list(self.img_loc_B[:, 1])

        return table


@dataclass
class SynchedStereoPointsPacket:
    sync_index: int
    stereo_points_packets: dict


    @property
    def pairs(self):
        return list(self.stereo_points_packets.keys())

    def to_table(self):
        pass


if __name__ == "__main__":
    from pyxy3d.recording.recorded_stream import RecordedStreamPool
    from pyxy3d.calibration.charuco import Charuco

    logger.setLevel(logging.DEBUG)
    from pyxy3d import __root__

    session_directory = Path(__root__, "tests", "5_cameras", "recording")
    csv_output = Path(session_directory, "paired_point_data.csv")

    ports = [0, 1, 2, 3, 4]

    tracker = Charuco(
        4, 5, 11, 8.5, aruco_scale=0.75, square_size_overide_cm=5.25, inverted=True
    )

    recorded_stream_pool = RecordedStreamPool(ports, session_directory, charuco=tracker)
    syncr = Synchronizer(recorded_stream_pool.streams, fps_target=200)
    recorded_stream_pool.play_videos()

    point_stream = StereoPointsBuilder(synchronizer=syncr, csv_output_path=csv_output)

    # I think that EOF needs to propogate up
    while not point_stream.frames_complete:
        points_packet = point_stream.out_q.get()

        # print("--------------------------------------")
        # print(points_packet)

    print("Saving data....")
    # save_data = pd.DataFrame(point_stream.tidy_output)
    # save_data.to_csv(csv_output)



import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import sys
from pathlib import Path
import numpy as np
import pandas as pd

import pyqtgraph.opengl as gl

from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtWidgets import (
    QApplication,
    QSlider,
    QVBoxLayout,
    QWidget,
)
from pyxy3d.session.session import Session
from pyxy3d.gui.vizualize.camera_mesh import CameraMesh, mesh_from_camera
from pyxy3d.cameras.camera_array import CameraArray


class PlaybackTriangulationWidget(QWidget):
    def __init__(self, camera_array: CameraArray, xyz_history_path: Path = None):
        super(PlaybackTriangulationWidget, self).__init__()

        self.camera_array = camera_array

        self.visualizer = TriangulationVisualizer(self.camera_array)
        # self.visualizer.scene.show()
        self.slider = QSlider(Qt.Orientation.Horizontal)

        # these defaults mean nothing right now without xyz data. Just placeholders

        self.setMinimumSize(500, 500)

        self.place_widgets()
        self.connect_widgets()
        if xyz_history_path is not None:
            xyz_history = pd.read_csv(xyz_history_path)
            self.set_xyz(xyz_history)

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.visualizer.scene)
        self.layout().addWidget(self.slider)

    def connect_widgets(self):
        self.slider.valueChanged.connect(self.visualizer.display_points)

    def set_xyz(self, xyz: pd.DataFrame):
        # self.xyz_history = pd.read_csv(xyz_history)
        self.visualizer.set_xyz(xyz)
        if xyz is not None:
            self.slider.setMinimum(self.visualizer.min_sync_index)
            self.slider.setMaximum(self.visualizer.max_sync_index)
        else:
            self.slider.setMinimum(0)
            self.slider.setMaximum(100)

    def update_camera_array(self, camera_array:CameraArray):
        self.visualizer.update_camera_array(camera_array)
class TriangulationVisualizer:
    """
    Can except either a single camera array or a capture volume that includes
    point_estimates. If a capture volume is supplied, point positions can
    be played back.
    """

    def __init__(self, camera_array: CameraArray):
        self.camera_array = camera_array
        self.build_scene()

    def build_scene(self):
        # constuct a scene if not yet there
        if hasattr(self, "scene"):
            logger.info("Clearing scene in capture volume visualizer")
            self.scene.clear()
        else:
            logger.info("Creating initial scene in capture volume visualizer")
            self.scene = gl.GLViewWidget()
            self.scene.setCameraPosition(distance=4)  # the scene camera, not a real Camera
        axis = gl.GLAxisItem()
        self.scene.addItem(axis)

        if self.camera_array.extrinsics_calibrated:
            self.meshes = {}
            for port, cam in self.camera_array.cameras.items():
                mesh: CameraMesh = mesh_from_camera(cam)
                self.meshes[port] = mesh
                self.scene.addItem(mesh)

        self.scatter = gl.GLScatterPlotItem(
            pos=np.array([0, 0, 0]),
            color=[1, 1, 1, 1],
            size=0.01,
            pxMode=False,
        )
        self.scene.addItem(self.scatter)
        self.scatter.setData(pos=None)
    
    def update_camera_array(self, camera_array:CameraArray):
        self.camera_array = camera_array
        self.build_scene()


    def set_xyz(self, xyz_history: pd.DataFrame):
        logger.info(f"Updating xyz history in playback widget")
        self.xyz_history = xyz_history

        if self.xyz_history is not None:
            self.sync_indices = self.xyz_history["sync_index"]
            self.min_sync_index = np.min(self.sync_indices)
            self.max_sync_index = np.max(self.sync_indices)
            self.sync_index = self.min_sync_index

            x_coord = self.xyz_history["x_coord"]
            y_coord = self.xyz_history["y_coord"]
            z_coord = self.xyz_history["z_coord"]
            self.xyz_coord = np.vstack([x_coord, y_coord, z_coord]).T

        else:
            self.xyz_coord = None
            self.sync_index=0
            # self.scatter.setData(pos=None)

        self.display_points(self.sync_index)

    def display_points(self, sync_index):
        """
        sync_index is provided from the dialog and linked to the slider
        it is initially set to the minimum viable sync index
        """

        if self.xyz_coord is not None:
            self.sync_index = sync_index

            current_sync_index_flag = self.sync_indices == self.sync_index
            self.points = self.xyz_coord[current_sync_index_flag]
            logger.debug(f"Displaying xyz points for sync index {sync_index}")
            self.scatter.setData(pos=self.points)

        else: 
            self.scatter.setData(pos=None)


# Environment for managing all created objects and the primary interface for the GUI.
import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

from PyQt6.QtCore import QObject, pyqtSignal
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path
from time import sleep
from enum import Enum, auto

from pyxy3d.trackers.charuco_tracker import CharucoTracker
from pyxy3d.calibration.monocalibrator import MonoCalibrator
from pyxy3d.cameras.camera import Camera
from pyxy3d.cameras.synchronizer import Synchronizer
from pyxy3d.cameras.camera_array_initializer import CameraArrayInitializer
from pyxy3d.interface import Tracker

from pyxy3d.calibration.stereocalibrator import StereoCalibrator
from pyxy3d.calibration.capture_volume.point_estimates import PointEstimates
from pyxy3d.calibration.capture_volume.capture_volume import CaptureVolume
from pyxy3d.calibration.capture_volume.quality_controller import QualityController

from pyxy3d.cameras.camera_array import CameraArray
from pyxy3d.calibration.capture_volume.helper_functions.get_point_estimates import (
    get_point_estimates,
)
from pyxy3d.configurator import Configurator
from pyxy3d.cameras.live_stream import LiveStream
from pyxy3d.recording.video_recorder import VideoRecorder

# %%
MAX_CAMERA_PORT_CHECK = 10
FILTERED_FRACTION = 0.05  # by default, 5% of image points with highest reprojection error are filtered out during calibration


class SessionMode(Enum):
    """ """

    Charuco = auto()
    IntrinsicCalibration = auto()
    ExtrinsicCalibration = auto()
    CaptureVolumeOrigin = auto()
    Recording = auto()
    PostProcessing = auto()


class Session(QObject):
    stream_tools_loaded_signal = pyqtSignal()
    stream_tools_disconnected_signal = pyqtSignal()
    unlock_postprocessing = pyqtSignal()
    recording_complete_signal = pyqtSignal()        

    mode_change_success = pyqtSignal(SessionMode)

    def __init__(self, config: Configurator):
        super(Session, self).__init__()
        self.config = config
        # self.folder = PurePath(directory).name
        self.path = self.config.session_path

        # this will not have anything to start, but the path
        # will be set
        self.extrinsic_calibration_xy = Path(
            self.path, "calibration", "extrinsic", "xy.csv"
        )

        # dictionaries of streaming related objects. key = port
        self.cameras = {}
        self.streams = {}
        self.stream_tools_in_process = False
        self.stream_tools_loaded = False
        # dictionaries of calibration related objects.
        self.monocalibrators = {}  # key = port
        self.active_monocalibrator = None

        # load fps for various modes
        self.fps_recording = self.config.get_fps_recording()
        self.fps_extrinsic_calibration = self.config.get_fps_extrinsic_calibration()
        self.fps_intrinsic_calibration = self.config.get_fps_intrinsic_calibration()
        self.wait_time_extrinsic = self.config.get_extrinsic_wait_time()
        self.wait_time_intrinsic = self.config.get_intrinsic_wait_time()

        self.is_recording = False

        self.charuco = self.config.get_charuco()
        self.charuco_tracker = CharucoTracker(self.charuco)
        self.mode = SessionMode.Charuco  # default mode of session

    def disconnect_cameras(self):
        """
        Shut down the streams, close the camera captures and delete the monocalibrators and synchronizer
        """
        for port, stream in self.streams.items():
            stream.stop_event.set()
        self.streams = {}
        for port, cam in self.cameras.items():
            cam.disconnect()
        self.cameras = {}
        for port, monocal in self.monocalibrators.items():
            monocal.stop_event.set()
        self.monocalibrators = {}
        self.synchronizer.stop_event.set()
        self.synchronizer = None

        self.stream_tools_disconnected_signal.emit()

    def is_camera_setup_eligible(self):
        if len(self.cameras) > 0:
            eligible = True
        else:
            eligible = False
        return eligible

    def is_extrinsic_calibration_eligible(self):
        # assume it is and prove if it's not
        self.config.refresh_from_toml()
        self.camera_array = self.config.get_camera_array()
        eligible = True
        for port, cam in self.camera_array.cameras.items():
            if cam.matrix is None or cam.distortions is None:
                eligible = False

        return eligible

    def is_capture_volume_eligible(self):
        """
        if all cameras that are not ignored have rotation and translation not None
        and there is a capture volume loaded up, then you can launch direct
        into the capture volume widget rather than the extrinsic calibration widget
        """

        # update current data
        self.config.refresh_from_toml()
        self.camera_array = self.config.get_camera_array()

        # assume it is and prove if it's not
        eligible = True
        if len(self.camera_array.cameras) < 2:
            eligible = False

        for port, cam in self.camera_array.cameras.items():
            if cam.rotation is None or cam.translation is None:
                eligible = False
                logger.info(
                    f"Failed capture volume eligibility due to camera {port}: {cam.__dict__}"
                )

        logger.info(
            f"Eligible to load capture volume? (i.e. fully calibrated extrinsics): {eligible}"
        )

        return eligible

    def is_recording_eligible(self):
        """
        Used to determine if the Record Button is enabled

        """

        # update current data
        self.config.refresh_from_toml()
        self.camera_array = self.config.get_camera_array()

        # assume true and prove otherwise
        eligible = True
        if len(self.camera_array.cameras) < 2:
            eligible = False

        for port, camera in self.camera_array.cameras.items():
            if camera.ignore == False and (
                camera.rotation is None or camera.translation is None
            ):
                eligible = False
                logger.info(
                    f"Camera array is not fully calibrated because camera {port} lacks extrinsics"
                )
        if eligible:
            eligible = True
        else:
            eligible = False

        return eligible

    def is_post_processing_eligible(self):
        """
        Post processing can only be performed if recordings (mp4 files) exist and extrinsics 
        (config.toml) are calibrated in the 'record' directory
        """
        #assume false and prove otherwise
        eligible = False
        for child in self.path.iterdir():
            if child.is_dir():
                mp4_files = list(child.glob('*.mp4'))
                config_file = child / 'config.toml'
                if mp4_files and config_file.exists():
                    eligible=True

        return eligible

    def set_mode(self, mode: SessionMode):
        """
        Via this method, the frame reading behavior will be changed by the GUI. If some properties are
        not available (i.e. synchronizer) they will be created
        """
        self.mode = mode
        self.update_streams_fps()

        match self.mode:
            case SessionMode.Charuco:
                if self.stream_tools_loaded:
                    self.synchronizer.unsubscribe_from_streams()
                    self.pause_all_monocalibrators()

            case SessionMode.PostProcessing:
                if self.stream_tools_loaded:
                    self.synchronizer.unsubscribe_from_streams()
                    self.pause_all_monocalibrators()

            case SessionMode.IntrinsicCalibration:
                # update in case something has changed
                self.charuco_tracker = CharucoTracker(self.charuco)

                if not self.stream_tools_loaded:
                    self.load_stream_tools()
                self.synchronizer.unsubscribe_from_streams()
                self.pause_all_monocalibrators()
                self.set_streams_charuco()
                self.set_streams_tracking(True)
                self.activate_monocalibrator(self.active_monocalibrator)

            case SessionMode.ExtrinsicCalibration:
                # update in case something has changed
                self.charuco_tracker = CharucoTracker(self.charuco)
                if not self.stream_tools_loaded:
                    self.load_stream_tools()

                self.pause_all_monocalibrators()
                self.set_streams_charuco()
                self.set_streams_tracking(True)
                
                self.synchronizer.subscribe_to_streams()

            case SessionMode.CaptureVolumeOrigin:
                if self.stream_tools_loaded:
                    self.pause_all_monocalibrators()
                    self.synchronizer.unsubscribe_from_streams()

            case SessionMode.Recording:
                logger.info("Attempting to set recording mode")
                if not self.stream_tools_loaded:
                    logger.info("Stream tools not loaded, so loading them up...")
                    self.load_stream_tools()

                logger.info("Pausing monocals to enter recording mode")
                self.pause_all_monocalibrators()
                
                logger.info("Stop tracking for recording mode")
                self.set_streams_tracking(False)
                logger.info("Update stream fps to recording fps")
                self.update_streams_fps()
                
                logger.info("Subscribe synchronizer to streams so video recorder can manage")
                self.synchronizer.subscribe_to_streams()

    def set_active_mode_fps(self, fps_target: int):
        """
        Updates the FPS used by the currently active session mode
        This update includes the config.toml
        """
        match self.mode:
            case SessionMode.Charuco:
                pass
            case SessionMode.PostProcessing:
                pass
            case SessionMode.IntrinsicCalibration:
                self.fps_intrinsic_calibration = fps_target
                self.config.save_fps_intrinsic_calibration(fps_target)
            case SessionMode.ExtrinsicCalibration:
                self.fps_extrinsic_calibration = fps_target
                self.config.save_fps_extrinsic_calibration(fps_target)
            case SessionMode.Recording:
                self.fps_recording = fps_target
                self.config.save_fps_recording(fps_target)

        self.update_streams_fps()

    def get_active_mode_fps(self) -> int:
        fps = None
        match self.mode:
            case SessionMode.Charuco:
                pass
            case SessionMode.PostProcessing:
                pass
            case SessionMode.IntrinsicCalibration:
                fps = self.fps_intrinsic_calibration
            case SessionMode.ExtrinsicCalibration:
                fps = self.fps_extrinsic_calibration
            case SessionMode.Recording:
                fps = self.fps_recording
        return fps

    def update_streams_fps(self):
        active_mode_fps = self.get_active_mode_fps()
        if active_mode_fps is not None:
            for port, stream in self.streams.items():
                stream.set_fps_target(active_mode_fps)

    def set_streams_tracking(self, tracking_on: bool):
        for port, stream in self.streams.items():
            stream.set_tracking_on(tracking_on)

    def set_streams_charuco(self):
        logger.info("updating charuco in case necessary")
        self.charuco_tracker = CharucoTracker(self.charuco)
        for port, stream in self.streams.items():
            stream.update_tracker(self.charuco_tracker)

    def pause_synchronizer(self):
        logger.info("pausing synchronizer")
        self.synchronizer.unsubscribe_from_streams()

    def unpause_synchronizer(self):
        self.synchronizer.subscribe_to_streams()
        self.synchronizer.set_stream_fps(self.synchronizer.fps_target)

    def get_configured_camera_count(self):
        count = 0
        for key, params in self.config.dict.copy().items():
            if key.startswith("cam"):
                count += 1
        return count

    def _find_cameras(self):
        """
        Called by load_streams in the event that no cameras are returned by the configurator...
        Will populate self.cameras using multiple threads
        """

        def add_cam(port):
            try:
                logger.info(f"Trying port {port}")
                cam = Camera(port)
                logger.info(f"Success at port {port}")
                self.cameras[port] = cam
                self.config.save_camera(cam)
                logger.info(
                    f"Loading stream at port {port} with charuco tracker for calibration"
                )
                self.streams[port] = LiveStream(
                    cam, tracker=CharucoTracker(self.charuco)
                )
            except:
                logger.info(f"No camera at port {port}")

        with ThreadPoolExecutor() as executor:
            for i in range(0, MAX_CAMERA_PORT_CHECK):
                if i in self.cameras.keys():
                    # don't try to connect to an already connected camera
                    pass
                else:
                    executor.submit(add_cam, i)

        # remove potential stereocalibration data to start fresh
        for key in self.config.dict.copy().keys():
            if key.startswith("stereo"):
                del self.config.dict[key]

    def load_stream_tools(self):
        """
        Connects to stored cameras and creates streams with provided tracking
        Because these streams are available, the synchronizer can then be initialized
        """
        self.stream_tools_in_process = True
        # don't bother loading cameras until you load the streams
        self.cameras = self.config.get_cameras()

        if len(self.cameras) == 0:
            self._find_cameras()

        for port, cam in self.cameras.items():
            if port in self.streams.keys():
                pass  # only add if not added yet
            else:
                logger.info(f"Loading Stream for port {port}")
                self.streams[port] = LiveStream(cam, tracker=self.charuco_tracker)

        self._adjust_resolutions()
        self._load_monocalibrators()

        # set something as a default value just to get started
        self.active_monocalibrator = min(self.monocalibrators.keys())

        self.synchronizer = Synchronizer(
            self.streams
        )  # defaults to stream default fps of 6

        # recording widget becomes available when synchronizer is created
        self.stream_tools_loaded = True
        self.stream_tools_in_process = False

        logger.info("Pausing stream tools since default loads to charuco")
        self.pause_all_monocalibrators()
        self.pause_synchronizer()
        
        logger.info(f"Signalling successful loading of stream tools")
        self.stream_tools_loaded_signal.emit()

    def _load_monocalibrators(self):
        for port, cam in self.cameras.items():
            if port in self.monocalibrators.keys():
                logger.info(
                    f"Skipping over monocalibrator creation for port {port} because it already exists."
                )
                pass  # only add if not added yet
            else:
                logger.info(f"Loading Monocalibrator for port {port}")
                self.monocalibrators[port] = MonoCalibrator(self.streams[port])

    def activate_monocalibrator(self, active_port: int = None):
        """
        Used to make sure that only the active camera tab is reading frames during the intrinsic calibration process
        """

        if active_port is not None:
            logger.info(f"Setting session active monocalibrator to {active_port}")
            self.active_monocalibrator = active_port
        logger.info(
            f"Activate tracking on port {self.active_monocalibrator} and deactivate others"
        )
        self.monocalibrators[self.active_monocalibrator].subscribe_to_stream()

    def pause_all_monocalibrators(self):
        """
        used when not actively on the camera calibration tab
        or when silencing all in preparation for activating only one
        """
        logger.info(f"Pausing all monocalibrator looping...")
        for port, monocal in self.monocalibrators.items():
            monocal.unsubscribe_to_stream()

    def start_recording(
        self, destination_directory: Path, store_point_history: bool = False
    ):
        logger.info("Initiating recording...")
        destination_directory.mkdir(parents=True, exist_ok=True)

        self.video_recorder = VideoRecorder(self.synchronizer)
        self.video_recorder.start_recording(
            destination_directory, store_point_history=store_point_history
        )
        self.is_recording = True

    def stop_recording(self):
        logger.info("Stopping recording...")
        self.video_recorder.stop_recording()
        while self.video_recorder.recording:
            logger.info("Waiting for video recorder to save out data...")
            sleep(0.5)

        self.is_recording = False

        logger.info(f"Recording of frames is complete...signalling change in status")
        self.recording_complete_signal.emit()

        if self.is_post_processing_eligible():
            self.unlock_postprocessing.emit()

    def _adjust_resolutions(self):
        """Changes the camera resolution to the value in the configuration, as
        log as it is not configured for the default resolution"""

        def adjust_res_worker(port):
            stream = self.streams[port]
            size = self.config.dict[f"cam_{port}"]["size"]
            default_size = self.cameras[port].default_resolution

            if size[0] != default_size[0] or size[1] != default_size[1]:
                logger.info(
                    f"Beginning to change resolution at port {port} from {default_size[0:2]} to {size[0:2]}"
                )
                stream.change_resolution(size)
                logger.info(
                    f"Completed change of resolution at port {port} from {default_size[0:2]} to {size[0:2]}"
                )

        with ThreadPoolExecutor() as executor:
            for port in self.cameras.keys():
                executor.submit(adjust_res_worker, port)

    def load_estimated_capture_volume(self):
        """
        Following capture volume optimization via bundle adjustment, or alteration
        via a transform of the origin, the entire capture volume can be reloaded
        from the config data without needing to go through the steps

        """
        self.point_estimates = self.config.get_point_estimates()
        self.camera_array = self.config.get_camera_array()
        self.capture_volume = CaptureVolume(self.camera_array, self.point_estimates)
        # self.capture_volume.rmse = self.config["capture_volume"]["RMSE"]
        self.capture_volume.stage = self.config.dict["capture_volume"]["stage"]
        if "origin_sync_index" in self.config.dict["capture_volume"].keys():
            self.capture_volume.origin_sync_index = self.config.dict["capture_volume"][
                "origin_sync_index"
            ]

        # QC needed to get the corner distance accuracy within the GUI
        self.quality_controller = QualityController(
            self.capture_volume, charuco=self.charuco
        )

    def estimate_extrinsics(self):
        """
        This is where the camera array 6 DoF is set. Many, many things are happening
        here, but they are all necessary steps of the process so I didn't want to
        try to encapsulate any further
        """
        stereocalibrator = StereoCalibrator(
            self.config.toml_path, self.extrinsic_calibration_xy
        )
        stereocalibrator.stereo_calibrate_all(boards_sampled=10)

        self.camera_array: CameraArray = CameraArrayInitializer(
            self.config.toml_path
        ).get_best_camera_array()

        self.point_estimates: PointEstimates = get_point_estimates(
            self.camera_array, self.extrinsic_calibration_xy
        )

        self.capture_volume = CaptureVolume(self.camera_array, self.point_estimates)
        self.capture_volume.optimize()

        self.quality_controller = QualityController(self.capture_volume, self.charuco)

        logger.info(
            f"Removing the worst fitting {FILTERED_FRACTION*100} percent of points from the model"
        )
        self.quality_controller.filter_point_estimates(FILTERED_FRACTION)
        self.capture_volume.optimize()

        self.config.save_capture_volume(self.capture_volume)



import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import sys
import os
from PyQt6.QtWidgets import QApplication
from pathlib import Path

from pyxy3d.gui.calibration_widget import launch_extrinsic_calibration_widget
from pyxy3d.gui.recording_widget import launch_recording_widget
from pyxy3d.gui.main_widget import launch_main

def CLI_parser():
    if len(sys.argv) == 1:
        # logger.warn("No argument supplied")
        # logger.warn(
        #     "Please provide specific function to perform: calibrate, record, process"
        # )
        launch_main()

    if len(sys.argv) == 2:
        session_path = Path(os.getcwd())
        launch_widget = sys.argv[1]

        if launch_widget in ["calibrate", "cal", "-c"]:
            launch_extrinsic_calibration_widget(session_path)

        if launch_widget in ["record", "rec", "-r"]:
            launch_recording_widget(session_path)




import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import os
import sys
import shutil
import time
from pathlib import Path
from threading import Thread
from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QIcon, QAction
from PyQt6.QtWidgets import (
    QWidget,
    QApplication,
    QVBoxLayout,
    QHBoxLayout,
    QDockWidget,
    QFileDialog,
    QStackedWidget,
)

from pyxy3d.session.session import Session, SessionMode
from pyxy3d.gui.charuco_widget import CharucoWidget
from pyxy3d.gui.camera_config.intrinsic_calibration_widget import IntrinsicCalibrationWidget
from pyxy3d import __root__, __app_dir__
from pyxy3d.trackers.charuco_tracker import CharucoTracker
# from pyxy3d.gui.qt_logger import QtLogger
from pyxy3d.gui.extrinsic_calibration_widget import (
    ExtrinsicCalibrationWidget,
    MIN_THRESHOLD_FOR_EARLY_CALIBRATE,
)
from pyxy3d.gui.vizualize.calibration.capture_volume_widget import CaptureVolumeWidget
from pyxy3d.configurator import Configurator


class CalibrationWidget(QStackedWidget):
    cameras_connected = pyqtSignal()

    def __init__(self, session:Session):
        super().__init__()
        self.CAMS_IN_PROCESS = False

        self.setWindowTitle("Camera Calibration Wizard")
        self.setWindowIcon(QIcon(str(Path(__root__, "pyxy3d/gui/icons/pyxy_logo.svg"))))
        self.session_path = session.path
        self.session = session
        self.config = session.config

        # self.launch_session()
        logger.info("Creating charuco wizard session")
        self.wizard_charuco = CharucoWidget(self.session)
        
        self.wizard_charuco.navigation_bar.next_wizard_step_btn.clicked.connect(
            self.activate_camera_config
        )
        logger.info("Adding charuco wizard")
        self.session.set_mode(SessionMode.Charuco)
        self.addWidget(self.wizard_charuco)

        logger.info("Setting index to 2 to activate widget")
        self.setCurrentWidget(self.wizard_charuco)

        self.connect_widgets()

    def connect_widgets(self):
        # self.wizard_directory.launch_wizard_btn.clicked.connect(self.begin_wizard)
        # self.cameras_connected.connect(self.on_cameras_connect)
        pass



    def activate_camera_config(self):
        self.session.set_mode(SessionMode.IntrinsicCalibration)
        if not hasattr(self, "intrinsic_calibration_widget"):
            logger.info(f"No camera configuration yet...creating wizard")
            self.intrinsic_calibration_widget = IntrinsicCalibrationWidget(self.session)
            self.addWidget(self.intrinsic_calibration_widget)
            self.setCurrentWidget(self.intrinsic_calibration_widget)
            # self.intrinsic_calibration_widget.navigation_bar.back_btn.clicked.connect(
            #     self.activate_charuco_wizard
            # )
            # self.intrinsic_calibration_widget.navigation_bar.next_btn.clicked.connect(
            #     self.next_to_stereoframe
            # )
        else:
            logger.info("Camera config already exists; changing stack current index")
            # active_port = self.camera_wizard.camera_tabs.currentIndex()
            # self.session.active_monocalibrator = active_port
            self.setCurrentWidget(self.intrinsic_calibration_widget)
            self.session.activate_monocalibrator()

    ####################### STEP 2: Single Camera Calibration #################
    def activate_charuco_wizard(self):
        self.session.set_mode(SessionMode.Charuco)
        self.setCurrentWidget(self.wizard_charuco)

    def next_to_stereoframe(self):
        self.session.set_mode(SessionMode.ExtrinsicCalibration)

        if hasattr(self, "extrinsic_calibration_widget"):
            self.setCurrentWidget(self.extrinsic_calibration_widget)
            # self.stereoframe.deleteLater()
            # self.removeWidget(self.stereoframe)
        else: 
            self.extrinsic_calibration_widget = ExtrinsicCalibrationWidget(self.session)
            self.addWidget(self.extrinsic_calibration_widget)
            self.setCurrentWidget(self.extrinsic_calibration_widget)

            self.extrinsic_calibration_widget.navigation_bar.back_btn.clicked.connect(
                self.back_to_camera_config_wizard
            )

            self.extrinsic_calibration_widget.calibration_complete.connect(self.next_to_capture_volume)
            # self.stereoframe.calibration_initiated.connect(self.show_calibration_qt_logger)
            self.extrinsic_calibration_widget.terminate.connect(self.refresh_stereoframe)

    ###################### Stereocalibration  ######################################
    def refresh_stereoframe(self):
        logger.info("Set current widget to config temporarily")
        self.setCurrentWidget(self.intrinsic_calibration_widget)

        logger.info("Remove stereoframe")
        self.removeWidget(self.extrinsic_calibration_widget)
        self.extrinsic_calibration_widget.paired_frame_builder.unsubscribe_from_synchronizer()
        del self.extrinsic_calibration_widget

        logger.info("Create new stereoframe")
        self.launch_new_stereoframe()


    def back_to_camera_config_wizard(self):
        logger.info("Moving back to camera config from stereoframe")
        self.setCurrentWidget(self.intrinsic_calibration_widget)
        self.session.pause_synchronizer()

        self.extrinsic_calibration_widget.paired_frame_builder.unsubscribe_from_synchronizer()
        self.removeWidget(self.extrinsic_calibration_widget)
        del self.extrinsic_calibration_widget

        active_port = self.intrinsic_calibration_widget.camera_tabs.currentIndex()
        self.intrinsic_calibration_widget.camera_tabs.activate_current_tab(active_port)

    def next_to_capture_volume(self):
        logger.info("Creating Capture Volume widget")
        self.capture_volume = CaptureVolumeWidget(self.session)
        logger.info("Adding capture volume widget to main Wizard")
        self.addWidget(self.capture_volume)
        logger.info("Set current index to capture volume widget")
        self.setCurrentWidget(self.capture_volume)
        self.capture_volume.navigation_bar.back_btn.clicked.connect(
            self.back_to_stereo_frame
        )

        del self.qt_logger

    ################## Capture Volume ########################
    def back_to_stereo_frame(self):
        logger.info("Set current widget to config temporarily")
        self.setCurrentWidget(self.intrinsic_calibration_widget)

        logger.info("Remove stereoframe")
        self.removeWidget(self.extrinsic_calibration_widget)
        logger.info("Remove capture volume")
        self.removeWidget(self.capture_volume)
        del self.capture_volume
        self.extrinsic_calibration_widget.paired_frame_builder.unsubscribe_from_synchronizer()
        del self.extrinsic_calibration_widget

        logger.info("Create new stereoframe")
        self.launch_new_stereoframe()
        self.session.unpause_synchronizer()


def launch_extrinsic_calibration_widget(session_path:Path):
    config = Configurator(session_path)
    session = Session(config)
    session.set_mode(SessionMode.ExtrinsicCalibration)

    app = QApplication(sys.argv)
    window = CalibrationWidget(session)
    window.show()
    app.exec()





import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import os
import sys
import shutil
import time
from pathlib import Path
from threading import Thread
from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QIcon, QAction
from PyQt6.QtWidgets import (
    QWidget,
    QApplication,
    QVBoxLayout,
    QHBoxLayout,
    QDockWidget,
    QFileDialog,
    QStackedWidget,
)

from pyxy3d.session.session import Session, SessionMode
from pyxy3d.gui.charuco_widget import CharucoWidget
from pyxy3d.gui.camera_config.intrinsic_calibration_widget import IntrinsicCalibrationWidget
from pyxy3d import __root__, __app_dir__
from pyxy3d.trackers.charuco_tracker import CharucoTracker
# from pyxy3d.gui.qt_logger import QtLogger
from pyxy3d.gui.extrinsic_calibration_widget import (
    ExtrinsicCalibrationWidget,
    MIN_THRESHOLD_FOR_EARLY_CALIBRATE,
)
from pyxy3d.gui.vizualize.calibration.capture_volume_widget import CaptureVolumeWidget
from pyxy3d.configurator import Configurator


class CalibrateCaptureVolumeWidget(QStackedWidget):
    """
    A combination of the Extrinsic Calibration Widget and the Capture Volume Widget
    Allows the user to move back and forth between the two functions.
    """
    cameras_connected = pyqtSignal()

    def __init__(self, session:Session):
        super().__init__()
        self.CAMS_IN_PROCESS = False

        self.session = session

        if self.session.is_capture_volume_eligible():
            logger.info("able to load Capture Volume, proceeeding with load of capture volume widget")
            self.activate_capture_volume_widget()
        else:
            logger.info("Unable to load capture volume from config; load extrinsic calibration widget")
            self.activate_extrinsic_calibration_widget()

    ###################### Stereocalibration  ######################################
    def activate_extrinsic_calibration_widget(self):
        logger.info(f"Setting session mode to {SessionMode.ExtrinsicCalibration} from within subwidget")
        self.session.set_mode(SessionMode.ExtrinsicCalibration)

        if hasattr(self, "extrinsic_calibration_widget"):
            logger.info("Activate extrinsic calibration widget")
            self.extrinsic_calibration_widget.deleteLater()
            self.extrinsic_calibration_widget = None
            # self.setCurrentWidget(self.extrinsic_calibration_widget)

        logger.info("Create new extrinsic calibration widget")
        self.extrinsic_calibration_widget = ExtrinsicCalibrationWidget(self.session)
        self.addWidget(self.extrinsic_calibration_widget)
        self.setCurrentWidget(self.extrinsic_calibration_widget)
        self.extrinsic_calibration_widget.calibration_complete.connect(self.activate_capture_volume_widget)
        self.extrinsic_calibration_widget.update_btn_eligibility()
        
        # self.session.unpause_synchronizer()
        

    def activate_capture_volume_widget(self):

        logger.info(f"Setting session mode to {SessionMode.CaptureVolumeOrigin} from within subwidget")
        self.session.set_mode(SessionMode.CaptureVolumeOrigin)

        if hasattr(self, "capture_volume_widget"):
            logger.info("Set current index to capture volume widget")
            self.capture_volume_widget.deleteLater()
            self.capture_volume_widget = None            

        logger.info("Creating Capture Volume widget")
        self.capture_volume_widget = CaptureVolumeWidget(self.session)
        logger.info("Adding capture volume widget to main Wizard")
        self.addWidget(self.capture_volume_widget)
        self.setCurrentWidget(self.capture_volume_widget)

        self.capture_volume_widget.recalibrate_btn.clicked.connect(
            self.activate_extrinsic_calibration_widget
        )
        
        # this will be managed elsewhere. StreamTools may or may not 
        # be loaded ...
        # self.session.pause_synchronizer()
            




import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import sys
from pathlib import Path
from threading import Thread
import time

import cv2
from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtGui import QImage, QPixmap, QIcon
from PyQt6.QtWidgets import (
    QApplication,
    QWidget,
    QSpinBox,
    QComboBox,
    QCheckBox,
    QDialog,
    QGroupBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSlider,
    QVBoxLayout,
)

# Append main repo to top of path to allow import of backend
from pyxy3d.gui.camera_config.frame_emitter import FrameEmitter
from pyxy3d.calibration.monocalibrator import MonoCalibrator
from pyxy3d.cameras.camera import Camera
from pyxy3d.cameras.live_stream import LiveStream
from pyxy3d.session.session import Session
from pyxy3d.gui.camera_config.camera_summary_widget import SummaryWidget
from pyxy3d import __root__


class CameraConfigTab(QDialog):
    def __init__(self, session, port):
        super(CameraConfigTab, self).__init__()

        # set up variables for ease of reference
        self.session = session
        self.monocal = session.monocalibrators[port]
        self.port = port
        self.stream = self.monocal.stream
        self.camera = self.stream.camera

        # need frame emitter to create actual frames and track FPS/grid count
        App = QApplication.instance()
        DISPLAY_WIDTH = App.primaryScreen().size().width()
        DISPLAY_HEIGHT = App.primaryScreen().size().height()

        self.pixmap_edge = min(DISPLAY_WIDTH / 3, DISPLAY_HEIGHT / 3)
        self.frame_emitter = FrameEmitter(self.monocal, self.pixmap_edge)
        self.frame_emitter.start()

        self.setWindowTitle("Camera Configuration and Calibration")
        self.setContentsMargins(0, 0, 0, 0)

        ################### BUILD SUB WIDGETS #############################

        ###################################################################

        self.setLayout(QHBoxLayout())
        ##########################################################
        ###################### CALIBRATION  ################################
        self.calibrate_grp = CalibrationControls(
            self.session, self.port, self.frame_emitter
        )

        # self.build_calibrate_grp()
        self.layout().addWidget(self.calibrate_grp)

        self.frame_controls_layout = QVBoxLayout(self)
        self.layout().addLayout(self.frame_controls_layout)
        self.basic_frame_control = FrameControlWidget(
            self.session, self.port, self.frame_emitter
        )

        self.frame_controls_layout.addWidget(self.basic_frame_control)

        self.frame_controls_layout.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        self.frame_controls_layout.setContentsMargins(0, 0, 0, 0)

        self.advanced_controls_toggle = QPushButton("Advanced Settings")
        self.advanced_controls_toggle.setCheckable(True)
        self.advanced_controls_toggle.clicked.connect(self.toggle_advanced_controls)

        self.frame_controls_layout.addWidget(self.advanced_controls_toggle)

        self.advanced_controls = AdvancedControls(
            self.session, self.port, self.frame_emitter
        )
        self.advanced_controls.hide()
        self.frame_controls_layout.addWidget(self.advanced_controls)

        self.connect_widgets()

    def toggle_advanced_controls(self):
        if self.advanced_controls_toggle.isChecked():
            self.advanced_controls.show()
        else:
            self.advanced_controls.hide()

    def connect_widgets(self):
        self.basic_frame_control.resolution_combo.currentTextChanged.connect(
            self.calibrate_grp.clear_camera_calibration
        )
        self.calibrate_grp.signal_calibration_lock.connect(self.lock_resolution_combo)

    def lock_resolution_combo(self, lock: bool):
        # note, the is locking, so the signal meaning gets reversed when setting "enable"
        enabled = not lock
        self.basic_frame_control.resolution_combo.setEnabled(enabled)

    def save_camera(self):
        self.session.save_camera(self.port)


class CalibrationControls(QGroupBox):
    signal_calibration_lock = pyqtSignal(bool)
    calibration_change = pyqtSignal()

    def __init__(self, session: Session, port, frame_emitter: FrameEmitter):
        super(CalibrationControls, self).__init__("Calibration Summary")

        self.session: Session = session
        self.port = port
        self.monocal: MonoCalibrator = self.session.monocalibrators[port]
        self.stream: LiveStream = self.monocal.stream
        self.camera: Camera = self.stream.camera
        self.frame_emitter = frame_emitter
        self.setLayout(QVBoxLayout())

        self.place_widgets()
        self.connect_widgets()

    def place_widgets(self):
        self.start_stop_calibration_btn = QPushButton("&Collect Data")
        self.layout().addWidget(self.start_stop_calibration_btn)
        self.undistort_btn = QPushButton("Un&distort")
        self.layout().addWidget(self.undistort_btn)

        if self.camera.matrix is None and self.camera.distortions is None:
            self.undistort_btn.setEnabled(False)

        self.camera_summary = SummaryWidget(self.camera)
        self.layout().addWidget(self.camera_summary)

    def connect_widgets(self):
        self.start_stop_calibration_btn.clicked.connect(self.capture_control)
        self.undistort_btn.clicked.connect(self.undistort)
        self.calibration_change.connect(self.update_camera_data)

    def capture_control(self):
        """change to turn on/off"""

        if self.start_stop_calibration_btn.text() == "&Collect Data":
            self.signal_calibration_lock.emit(True)
            self.clear_camera_calibration()
            self.calibration_change.emit()
            self.monocal.capture_corners.set()
            self.undistort_btn.setEnabled(False)
            self.start_stop_calibration_btn.setText("&Calibrate")

        elif self.start_stop_calibration_btn.text() == "&Calibrate":
            self.signal_calibration_lock.emit(True)
            if len(self.monocal.all_ids) > 0:
                # stop the colletion of more data
                self.monocal.capture_corners.clear()
                self.calibrate()
            else:
                # no data collected
                self.monocal.capture_corners.clear()
                self.update_camera_data()
                self.start_stop_calibration_btn.setText("&Collect Data")

        elif self.start_stop_calibration_btn.text() == "Re-&Collect":
            self.signal_calibration_lock.emit(True)
            self.clear_camera_calibration()
            self.calibration_change.emit()
            self.monocal.initialize_grid_history()
            self.undistort_btn.setEnabled(False)
            self.monocal.capture_corners.set()
            self.start_stop_calibration_btn.setText("&Calibrate")

    def clear_camera_calibration(self):
        self.camera.matrix = None
        self.camera.distortions = None
        self.camera.error = None
        self.camera.grid_count = None
        self.session.config.save_camera(self.camera)
        # self.camera_summary.place_widgets()
        self.update_camera_data()
        self.undistort_btn.setEnabled(False)

    def calibrate(self):
        def wrker():
            self.start_stop_calibration_btn.setText("---processing---")
            self.start_stop_calibration_btn.setEnabled(False)

            self.monocal.calibrate()
            self.session.config.save_camera(self.camera)
            self.calibration_change.emit()

            self.undistort_btn.setEnabled(True)
            self.start_stop_calibration_btn.setText("Re-&Collect")
            self.start_stop_calibration_btn.setEnabled(True)
            self.signal_calibration_lock.emit(False)

        self.calib_thread = Thread(target=wrker, args=(), daemon=True)
        self.calib_thread.start()

    def undistort(self):
        if self.undistort_btn.text() == "Un&distort":
            self.signal_calibration_lock.emit(True)
            self.start_stop_calibration_btn.setEnabled(False)
            self.frame_emitter.undistort = True
            self.undistort_btn.setText("Revert &Distortion")

        elif self.undistort_btn.text() == "Revert &Distortion":
            self.start_stop_calibration_btn.setEnabled(True)
            self.frame_emitter.undistort = False
            self.undistort_btn.setText("Un&distort")
            self.signal_calibration_lock.emit(False)

    def update_camera_data(self):
        self.layout().removeWidget(self.camera_summary)
        self.camera_summary.deleteLater()
        self.camera_summary = None
        self.camera_summary = SummaryWidget(self.camera)
        self.layout().addWidget(self.camera_summary)


class AdvancedControls(QWidget):
    def __init__(self, session: Session, port, frame_emitter: FrameEmitter):
        super(AdvancedControls, self).__init__()
        self.session: Session = session
        self.port = port
        self.monocal: MonoCalibrator = self.session.monocalibrators[port]
        self.stream: LiveStream = self.monocal.stream
        self.camera: Camera = self.stream.camera
        self.frame_emitter = frame_emitter
        self.setLayout(QHBoxLayout())

        self.place_widgets()
        self.connect_widgets()

    def place_widgets(self):
        self.fps_grp = QGroupBox("FPS")
        self.layout().addWidget(self.fps_grp)
        self.fps_grp.setLayout(QHBoxLayout())

        logger.debug("Building FPS Control")
        self.fps_grp.layout().addWidget(QLabel("Target:"))

        self.frame_rate_spin = QSpinBox()
        self.frame_rate_spin.setValue(self.stream.fps)
        self.fps_grp.layout().addWidget(self.frame_rate_spin)

        self.fps_display = QLabel()
        self.fps_grp.layout().addWidget(self.fps_display)
        self.grid_grp = QGroupBox("Grid Collection")
        self.layout().addWidget(self.grid_grp)
        self.grid_grp.setLayout(QHBoxLayout())

        self.grid_grp.layout().addWidget(QLabel("Wait Time:"))

        self.wait_time_spin = QDoubleSpinBox()
        self.wait_time_spin.setValue(self.monocal.wait_time)
        self.wait_time_spin.setSingleStep(0.1)

        self.wait_time_spin.valueChanged.connect(self.on_wait_time_spin)
        self.grid_grp.layout().addWidget(self.wait_time_spin)

        self.ignore_box = QCheckBox("Ignore", self)
        self.layout().addWidget(self.ignore_box)

    def connect_widgets(self):
        self.frame_rate_spin.valueChanged.connect(self.on_frame_rate_spin)

        def FPSUpdateSlot(fps):
            if self.monocal.camera.capture.isOpened():
                # rounding to nearest integer should be close enough for our purposes
                self.fps_display.setText("Actual: " + str(round(fps, 1)))
            else:
                self.fps_display.setText("reconnecting to camera...")

        self.frame_emitter.FPSBroadcast.connect(FPSUpdateSlot)
        self.ignore_box.stateChanged.connect(self.ignore_cam)

    def ignore_cam(self, signal):
        if signal == 0:  # not checked
            logger.info(f"Don't ignore camera at port {self.port}")
            self.camera.ignore = False
        else:  # value of checkState() might be 2?
            logger.info(f"Ignore camera at port {self.port}")
            self.camera.ignore = True
        self.session.config.save_camera(self.camera)

    def on_wait_time_spin(self, wait_time):
        self.session.wait_time_intrinsic = wait_time
        self.session.config.save_intrinsic_wait_time(wait_time)

    def on_frame_rate_spin(self, fps_rate):
        self.session.set_active_mode_fps(fps_rate)
        logger.info(f"Changing monocalibrator frame rate for port{self.port}")

    def FPSUpdateSlot(self, fps):
        if self.monocal.camera.capture.isOpened():
            # rounding to nearest integer should be close enough for our purposes
            self.fps_display.setText("Actual: " + str(round(fps, 1)))
        else:
            self.fps_display.setText("reconnecting to camera...")


class FrameControlWidget(QWidget):
    def __init__(self, session: Session, port, frame_emitter: FrameEmitter):
        super(FrameControlWidget, self).__init__()
        self.session: Session = session
        self.monocal: MonoCalibrator = session.monocalibrators[port]
        self.port = port
        self.camera: Camera = self.monocal.stream.camera
        self.frame_emitter = frame_emitter

        self.place_widgets()
        self.connect_widgets()

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        ###### FRAME ####################################
        # return a QLabel that is linked to the constantly changing image
        self.frame_display = QLabel()
        self.layout().addWidget(self.frame_display)

        self.rotation_resolution_hbox = QHBoxLayout()
        self.layout().addLayout(self.rotation_resolution_hbox)
        ###################### Rotation Buttons  ###################################
        # icons from https://iconoir.com
        self.cw_rotation_btn = QPushButton(
            QIcon(str(Path(__root__, "pyxy3d/gui/icons/rotate-camera-right.svg"))), ""
        )
        self.cw_rotation_btn.setMaximumSize(35, 35)
        self.ccw_rotation_btn = QPushButton(
            QIcon(str(Path(__root__, "pyxy3d/gui/icons/rotate-camera-left.svg"))), ""
        )

        self.ccw_rotation_btn.setMaximumSize(35, 35)
        self.rotation_resolution_hbox.addWidget(self.cw_rotation_btn)
        self.rotation_resolution_hbox.addWidget(self.ccw_rotation_btn)

        ##################### RESOLUTION DROP DOWN ##############################
        self.resolution_combo = QComboBox()

        resolutions_text = []
        for w, h in self.monocal.stream.camera.verified_resolutions:
            resolutions_text.append(f"{int(w)} x {int(h)}")

        self.resolution_combo.addItems(resolutions_text)
        self.resolution_combo.setMaximumSize(100, 35)

        w, h = self.monocal.camera.size
        self.resolution_combo.setCurrentText(f"{int(w)} x {int(h)}")

        self.rotation_resolution_hbox.addWidget(self.resolution_combo)

        ######################## EXPOSURE BOX #############################

        self.exposure_hbox = QHBoxLayout()
        self.layout().addLayout(self.exposure_hbox)

        # construct a horizontal widget with label: slider: value display
        self.exposure_label = QLabel("Exposure")
        self.exposure_label.setAlignment(Qt.AlignmentFlag.AlignRight)

        self.exp_slider = QSlider(Qt.Orientation.Horizontal)
        self.exp_slider.setRange(-10, 0)
        self.exp_slider.setSliderPosition(int(self.monocal.camera.exposure))
        self.exp_slider.setPageStep(1)
        self.exp_slider.setSingleStep(1)
        self.exp_slider.setMaximumWidth(200)
        self.exposure_number = QLabel()
        self.exposure_number.setText(str(int(self.monocal.camera.exposure)))

        self.exposure_hbox.addWidget(self.exposure_label)
        self.exposure_hbox.addWidget(self.exp_slider)
        self.exposure_hbox.addWidget(self.exposure_number)

        self.exposure_hbox.setAlignment(Qt.AlignmentFlag.AlignCenter)

    def connect_widgets(self):
        def ImageUpdateSlot(QPixmap):
            self.frame_display.setPixmap(QPixmap)

        self.frame_emitter.ImageBroadcast.connect(ImageUpdateSlot)

        # Counter Clockwise rotation called because the display image is flipped
        self.cw_rotation_btn.clicked.connect(self.monocal.camera.rotate_CCW)
        self.cw_rotation_btn.clicked.connect(self.save_camera)

        # Clockwise rotation called because the display image is flipped
        self.ccw_rotation_btn.clicked.connect(self.monocal.camera.rotate_CW)
        self.ccw_rotation_btn.clicked.connect(self.save_camera)

        # resolution combo box
        self.resolution_combo.currentTextChanged.connect(self.change_resolution)

        # exposure slider
        self.exp_slider.valueChanged.connect(self.update_exposure)
        # self.ignore_box.stateChanged.connect(self.ignore_cam)

    def save_camera(self):
        # normally wouldn't bother with a one-liner function, but it makes connecting
        # to the signal more straightforward
        self.session.config.save_camera(self.camera)


    def update_exposure(self, exp):
        self.monocal.camera.exposure = exp
        self.exposure_number.setText(str(exp))
        self.save_camera()

    def change_resolution(self, res_text):
        # call the cam_cap widget to change the resolution, but do it in a
        # thread so that it doesn't halt your progress

        w, h = res_text.split("x")
        w, h = int(w), int(h)
        new_res = (w, h)
        logger.info(f"Attempting to change resolution of camera at port {self.port}")

        def change_res_worker(new_res):
            self.monocal.stream.change_resolution(new_res)

            # clear out now irrelevant params
            self.camera.matrix = None
            self.camera.distortions = None
            self.camera.error = None
            self.camera.grid_count = None
            self.save_camera()

        self.change_res_thread = Thread(
            target=change_res_worker,
            args=(new_res,),
            daemon=True,
        )
        self.change_res_thread.start()




from dataclasses import dataclass
import numpy as np
from numba.typed import List
from queue import Queue
from abc import ABC, abstractmethod
import cv2


@dataclass(frozen=True,slots=True)
class PointPacket:
    """
    This will be the primary return value of the Tracker Protocol
    A calleable that receives an image frame and returns a point_packet
    """

    point_id: np.ndarray = (
        None  # unique point id that aligns with Tracker.get_point_names()
    )
    img_loc: np.ndarray = None  # x,y position of tracked point
    obj_loc: np.ndarray = (
        None  # x,y,z in object frame of reference; primarily for calibration
    )
    confidence: np.ndarray = (
        None  # may be available in some trackers..include for downstream
    )

    @property
    def obj_loc_list(self)->List[List]:
        """
        if obj loc is none, printing data to a table format will be undermined
        this creates a list of length len(point_id) that is empty so that a pd.dataframe
        can be constructed from it.
        """
        if self.obj_loc is not None:
            obj_loc_x = self.obj_loc[:,0].tolist()
            obj_loc_y = self.obj_loc[:,1].tolist()
        else:
            length = len(self.point_id)
            obj_loc_x = [None]*length 
            obj_loc_y = [None]*length 

        return [obj_loc_x,obj_loc_y]

class Tracker(ABC):
    @abstractmethod
    def get_points(self, frame: np.ndarray, port:int, rotation_count:int) -> PointPacket:
        pass

    @property
    def name(self)->str:
        """
        returns the tracker name in the same format as used by TrackerEnum
        Used for file naming creation
        """
        pass
        
    @abstractmethod
    def get_point_name(self, point_id:int) -> str:
        """
        Used for saving out data with sensible headers
        """
        pass

    def get_connected_points(self):
        """
        used for drawing purposes elsewhere. Specify which
        points (if any) should have a line connecting them

        """
        pass


    @abstractmethod
    def draw_instructions(self, point_id:int) ->dict:
        pass
    
    
    

class Stream(ABC):
    """
    As much an exercise in better understanding ABC as it is anything...
    """

    @abstractmethod
    def subscribe(self, queue: Queue):
        pass

    @abstractmethod
    def unsubscribe(self, queue: Queue):
        pass

    @abstractmethod
    def set_tracking_on(self, track: bool):
        pass
    
    @abstractmethod
    def set_fps_target(self, fps_target:int):
        pass
    
    @abstractmethod
    def set_fps_target(self, fps_target:int):
        pass

    @abstractmethod
    def process_frames(self):
        pass


@dataclass(slots=True)
class FramePacket:
    """
    Holds the data for a single frame from a camera, including the frame itself,
    the frame time and the points if they were generated
    """

    port: int
    frame_time: float
    frame: np.ndarray
    frame_index: int = None
    points: PointPacket = None
    draw_instructions: callable = None

    def to_tidy_table(self, sync_index)->dict:
        """
        Returns a dictionary of lists where each list is as long as the
        number of points identified on the frame; 
        used for creating csv output via pandas
        """
        if self.points is not None:
            point_count = len(self.points.point_id)
            if point_count > 0:
                table = {
                    "sync_index": [sync_index] * point_count,
                    "port": [self.port] * point_count,
                    "frame_index": [self.frame_index] * point_count,
                    "frame_time": [self.frame_time] * point_count,
                    "point_id": self.points.point_id.tolist(),
                    "img_loc_x": self.points.img_loc[:, 0].tolist(),
                    "img_loc_y": self.points.img_loc[:, 1].tolist(),
                    "obj_loc_x": self.points.obj_loc_list[0],
                    "obj_loc_y": self.points.obj_loc_list[1]
                }
            else:
                table = None
        else:
            table = None
        return table

    @property
    def frame_with_points(self):
        if self.points is not None:
            drawn_frame = self.frame.copy()
            ids = self.points.point_id
            locs = self.points.img_loc
            for _id, coord in zip(ids, locs):
                x = round(coord[0])
                y = round(coord[1])

                # draw instructions are a method of Tracker object
                params = self.draw_instructions(_id)
                cv2.circle(drawn_frame, (x, y),params["radius"], params["color"], params["thickness"])
        else:
            drawn_frame = self.frame

        return drawn_frame


@dataclass(frozen=True, slots=True)
class SyncPacket:
    """
    SyncPacket holds syncronized frame packets.
    """

    sync_index: int
    frame_packets: dict

    @property
    def triangulation_inputs(self):
        """
        returns three key items used by the triangulation functions
            cameras: a list of the camera ids associated with each reported 2d point
            point_ids: the point id associated with each 2d point
            img_xy: the 2d image points themselves

        """
        cameras = []
        point_ids = []
        img_xy = []

        for port, packet in self.frame_packets.items():
            if packet is not None and packet.points is not None:
                cameras.extend([port] * len(packet.points.point_id))
                point_ids.extend(packet.points.point_id.tolist())
                img_xy.extend(packet.points.img_loc.tolist())

        return cameras, point_ids, img_xy

    @property
    def dropped(self):
        """
        convencience method to ease tracking of dropped frame rate within the synchronizer
        """
        temp_dict = {}
        for port, packet in self.frame_packets.items():
            if packet is None:
                temp_dict[port] = 1
            else:
                temp_dict[port] = 0
        return temp_dict
    
    @property
    def frame_packet_count(self):
        count = 0
        for port, packet in self.frame_packets.items():
            if packet is not None:
                 count+= 1
        return count
        


@dataclass
class XYZPacket:
    sync_index: int
    point_ids: List
    point_xyz: List



#%%

import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import sys
from pathlib import Path
import numpy as np
import pandas as pd
from queue import Queue
from threading import Thread
import pyqtgraph.opengl as gl

from PyQt6.QtCore import Qt, pyqtSignal
from PyQt6.QtWidgets import (
    QApplication,
    QSlider,
    QVBoxLayout,
    QWidget,
)
from pyxy3d.session.session import Session
from pyxy3d.gui.vizualize.camera_mesh import CameraMesh, mesh_from_camera
from pyxy3d.interface import XYZPacket
from pyxy3d.cameras.camera_array import CameraArray

class RealTimeTriangulationWidget(QWidget):
    def __init__(self, camera_array:CameraArray, xyz_packet_q:Queue):
        super(RealTimeTriangulationWidget, self).__init__()

        self.camera_array = camera_array
        self.xyz_packet_q = xyz_packet_q #place to receive xyz_packets from Triangulator or similar...

        self.visualizer = TriangulationVisualizer(self.camera_array)

        self.setMinimumSize(500, 500)

        self.place_widgets()
        self.connect_widgets()

        self.thread = Thread(target=self.process_incoming,args=[],daemon=True)
        self.thread.start()

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        self.layout().addWidget(self.visualizer.scene)

    def connect_widgets(self):
        pass
        # self.slider.valueChanged.connect(self.visualizer.display_points)

    def process_incoming(self):
        
        while True:
            xyz_packet = self.xyz_packet_q.get()
            self.visualizer.display_points(xyz_packet)
            

class TriangulationVisualizer:
    """
    Can except either a single camera array or a capture volume that includes
    point_estimates. If a capture volume is supplied, point positions can
    be played back.
    """

    def __init__(
        self, camera_array: CameraArray
    ):

        self.camera_array = camera_array

        # self.min_sync_index = np.min(self.sync_indices)
        # self.max_sync_index = np.max(self.sync_indices)
        # self.sync_index = self.min_sync_index

        # x_coord = self.xyz_history["x_coord"]
        # y_coord = self.xyz_history["y_coord"]
        # z_coord = self.xyz_history["z_coord"]
        # self.xyz_coord = np.vstack([x_coord,y_coord,z_coord]).T
        
        # constuct a scene
        self.scene = gl.GLViewWidget()
        self.scene.setCameraPosition(distance=4)  # the scene camera, not a real Camera

        axis = gl.GLAxisItem()
        self.scene.addItem(axis)

        # build meshes for all cameras
        self.meshes = {}
        for port, cam in self.camera_array.cameras.items():
            print(port)
            print(cam)
            mesh:CameraMesh = mesh_from_camera(cam)
            self.meshes[port] = mesh
            self.scene.addItem(mesh)

        self.scatter = gl.GLScatterPlotItem(
            pos=np.array([0, 0, 0]),
            color=[1, 1, 1, 1],
            size=0.01,
            pxMode=False,
        )
        self.scene.addItem(self.scatter)

        # self.display_points(self.sync_index)
                 
    def display_points(self, xyz_packet:XYZPacket):
        # self.sync_index = sync_index
        # current_sync_index_flag = self.sync_indices == self.sync_index
        # self.single_board_points = self.xyz_coord[current_sync_index_flag]
        # logger.info(f"Displaying xyz points for sync index {sync_index}")
        self.scatter.setData(pos=xyz_packet.point_xyz)

if __name__ == "__main__":

    from PyQt6.QtWidgets import QApplication
    import time
    from pyxy3d import __root__
    from pyxy3d.calibration.capture_volume.helper_functions.get_point_estimates import (
        get_point_estimates,
    )

    from pyxy3d.calibration.capture_volume.capture_volume import CaptureVolume
    
    test_sessions = [
        # Path(__root__, "dev", "sample_sessions", "post_triangulation"),
        Path(__root__, "dev", "sample_sessions", "triangulated_hands", )
    ]

    test_session_index = 0
    session_path = test_sessions[test_session_index]
    logger.info(f"Loading session {session_path}")
    session = Session(session_path)

    session.load_estimated_capture_volume()

    app = QApplication(sys.argv)
    
    xyz_queue = Queue()
    
    vizr_dialog = RealTimeTriangulationWidget(session.capture_volume.camera_array, xyz_queue)

#%%   
    def send_packets_from_history():
        xyz_history_path = Path(session_path,"xyz_history.csv")
        
        xyz_data = pd.read_csv(xyz_history_path)
        sync_indices = np.unique(xyz_data["sync_index"])
        sync_indices.sort()
        
        for sync_index in sync_indices:
            current_xyz = xyz_data[xyz_data["sync_index"]==sync_index] 
            points = current_xyz["point_id"].to_numpy()
            xyz = current_xyz[["x_coord", "y_coord", "z_coord"]].to_numpy()
            
            xyz_packet = XYZPacket(sync_index,points,xyz )
            xyz_queue.put(xyz_packet)
            time.sleep(.01)  
        logger.info("attempting to quit application")
        # vizr_dialog.close()
        # app.quit()
        # sys.exit()
         
    thread = Thread(target=send_packets_from_history,args=[],daemon=True)
    thread.start()        

    vizr_dialog.show()

    sys.exit(app.exec())




import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

from datetime import datetime
from pathlib import Path

import cv2
from PyQt6.QtCore import QSize, Qt, QThread, pyqtSignal
from PyQt6.QtGui import QFont, QIcon, QImage, QPixmap


class FrameEmitter(QThread):
    # establish signals from the frame that will be displayed in real time
    # within the GUI
    ImageBroadcast = pyqtSignal(QPixmap)

    FPSBroadcast = pyqtSignal(float)
    GridCountBroadcast = pyqtSignal(int)

    def __init__(self, monocalibrator, pixmap_edge_length=None):
        # pixmap_edge length is from the display window. Keep the display area
        # square to keep life simple.
        super(FrameEmitter, self).__init__()
        self.monocalibrator = monocalibrator
        self.pixmap_edge_length = pixmap_edge_length
        self.rotation_count = monocalibrator.camera.rotation_count
        self.undistort = False

    def run(self):
        self.ThreadActive = True

        while self.ThreadActive:
            # Grab a frame from the queue and broadcast to displays
            self.monocalibrator.grid_frame_ready_q.get()

            self.frame = self.monocalibrator.grid_frame
            self.apply_undistortion()
            self.frame = resize_to_square(self.frame)
            self.apply_rotation()

            image = self.cv2_to_qlabel(self.frame)
            pixmap = QPixmap.fromImage(image)

            if self.pixmap_edge_length:
                pixmap = pixmap.scaled(
                    int(self.pixmap_edge_length),
                    int(self.pixmap_edge_length),
                    Qt.AspectRatioMode.KeepAspectRatio,
                )
            self.ImageBroadcast.emit(pixmap)
            
            # moved to monocalibrator...delete if works well
            self.FPSBroadcast.emit(self.monocalibrator.stream.FPS_actual)
            self.GridCountBroadcast.emit(self.monocalibrator.grid_count)

    def stop(self):
        self.ThreadActive = False
        self.quit()

    def cv2_to_qlabel(self, frame):
        Image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)
        FlippedImage = cv2.flip(Image, 1)

        qt_frame = QImage(
            FlippedImage.data,
            FlippedImage.shape[1],
            FlippedImage.shape[0],
            QImage.Format.Format_RGB888,
        )
        return qt_frame

    def apply_rotation(self):
        # logger.debug("Applying Rotation")
        if self.monocalibrator.camera.rotation_count == 0:
            pass
        elif self.monocalibrator.camera.rotation_count in [1, -3]:
            self.frame = cv2.rotate(self.frame, cv2.ROTATE_90_CLOCKWISE)
        elif self.monocalibrator.camera.rotation_count in [2, -2]:
            self.frame = cv2.rotate(self.frame, cv2.ROTATE_180)
        elif self.monocalibrator.camera.rotation_count in [-1, 3]:
            self.frame = cv2.rotate(self.frame, cv2.ROTATE_90_COUNTERCLOCKWISE)

    def apply_undistortion(self):
        
        if self.undistort == True:  # and self.mono_cal.is_calibrated:
            self.frame = cv2.undistort(
                self.frame,
                self.monocalibrator.camera.matrix,
                self.monocalibrator.camera.distortions,
            )


def resize_to_square(frame):

    height = frame.shape[0]
    width = frame.shape[1]

    padded_size = max(height, width)

    height_pad = int((padded_size - height) / 2)
    width_pad = int((padded_size - width) / 2)
    pad_color = [0, 0, 0]

    frame = cv2.copyMakeBorder(
        frame,
        height_pad,
        height_pad,
        width_pad,
        width_pad,
        cv2.BORDER_CONSTANT,
        value=pad_color,
    )

    return frame

if __name__ == "__main__":
    pass

    # not much to look at here... go to camera_config_dialogue.py for test




import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

import sys
from pathlib import Path
from threading import Thread, Event
import time
from enum import Enum

import cv2
from PyQt6.QtCore import Qt, pyqtSignal, QThread
from PyQt6.QtGui import QImage, QPixmap, QIcon
from PyQt6.QtWidgets import (
    QApplication,
    QSizePolicy,
    QWidget,
    QSpinBox,
    QScrollArea,
    QComboBox,
    QCheckBox,
    QDialog,
    QGroupBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSlider,
    QVBoxLayout,
)

# Append main repo to top of path to allow import of backend
from pyxy3d.session.session import Session
from pyxy3d.gui.frame_builders.paired_frame_builder import PairedFrameBuilder
from pyxy3d.cameras.synchronizer import Synchronizer
from pyxy3d import __root__
from pyxy3d.gui.navigation_bars import NavigationBarNext

# the boards needed before a pair could be used to bridge pairs without common corners
MIN_THRESHOLD_FOR_EARLY_CALIBRATE = 5


class PossibleActions(Enum):
    CollectData = "Collect Data"
    Terminate = "Terminate"
    Calibrate = "Calibrate"

class ExtrinsicCalibrationWidget(QWidget):
    calibration_complete = pyqtSignal()
    calibration_initiated = pyqtSignal()
    terminate = pyqtSignal()
     
    def __init__(self,session:Session):

        super(ExtrinsicCalibrationWidget, self).__init__()
        self.session = session
        self.synchronizer:Synchronizer = self.session.synchronizer

        self.paired_frame_builder = PairedFrameBuilder(self.synchronizer, board_count_target=30)
        self.paired_frame_emitter = PairedFrameEmitter(self.paired_frame_builder)
        self.paired_frame_emitter.start()

        self.frame_rate_spin = QSpinBox()
        self.frame_rate_spin.setValue(self.synchronizer.fps_target)
        self.board_count_spin = QSpinBox()
        self.board_count_spin.setValue(self.paired_frame_builder.board_count_target)
        
        self.stereo_frame_display = QLabel()
        # self.navigation_bar = NavigationBarNext() 
        self.possible_action = PossibleActions.CollectData
        self.calibrate_collect_btn = QPushButton(self.possible_action.value)
        
        self.place_widgets()
        self.connect_widgets()        

        self.update_btn_eligibility()
    
    def shutdown_threads(self):
        """
        There may be some lingering threads running when the extrinsic calibrator loses focus
        This may be causing python to overload and pyqt to segfault during the calibration process
        if I've moved from the extrinsic calibration widget to a different one...
        """
        logger.info("Unsubscribe paired frame builder from sync notice")
        self.paired_frame_builder.unsubscribe_from_synchronizer()
        logger.info("signal paired frame emitter to stop collecting frames")
        self.paired_frame_emitter.keep_collecting.clear()

    def update_btn_eligibility(self):
        if self.session.is_extrinsic_calibration_eligible():
            self.calibrate_collect_btn.setEnabled(True)
        else:
            self.calibrate_collect_btn.setEnabled(False)
    

    def place_widgets(self):
        self.setLayout(QVBoxLayout())
        
        self.settings_group = QGroupBox("Settings")
        self.settings_group.setLayout(QHBoxLayout())
        self.settings_group.layout().addWidget(QLabel("Frame Rate:"))
        self.settings_group.layout().addWidget(self.frame_rate_spin)       
        self.settings_group.layout().addWidget(QLabel("Target Board Count:"))
        self.settings_group.layout().addWidget(self.board_count_spin)       

        self.layout().addWidget(self.settings_group)
        # self.layout().addWidget(self.calibrate_collect_btn)

        # scroll bar appears to not be working....
        self.scroll_area = QScrollArea()
        self.scroll_area.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOn)
        # self.scroll_area.setLayout(QVBoxLayout())
        self.layout().addWidget(self.scroll_area)

        self.scroll_area.setWidget(self.stereo_frame_display)
       
        self.layout().addWidget(self.calibrate_collect_btn)



    def connect_widgets(self):
        
        self.calibrate_collect_btn.clicked.connect(self.on_calibrate_collect_click)
        self.paired_frame_emitter.ImageBroadcast.connect(self.ImageUpdateSlot)
        self.paired_frame_emitter.possible_to_initialize_array.connect(self.enable_calibration)
        self.frame_rate_spin.valueChanged.connect(self.synchronizer.set_stream_fps)
        self.board_count_spin.valueChanged.connect(self.update_board_count_target)
        self.paired_frame_emitter.calibration_data_collected.connect(self.initiate_calibration)
    
    def update_board_count_target(self, target):
        self.paired_frame_builder.board_count_target = target
        
    def on_calibrate_collect_click(self):
        if self.possible_action == PossibleActions.CollectData:
            logger.info("Begin collecting calibration data")
            # by default, data saved to session folder
            self.paired_frame_builder.store_points.set()
            extrinsic_calibration_path = Path(self.session.path, "calibration", "extrinsic")
            self.session.start_recording(extrinsic_calibration_path,store_point_history=True)
            self.possible_action = PossibleActions.Terminate
            self.calibrate_collect_btn.setText(self.possible_action.value)
            self.calibrate_collect_btn.setEnabled(True)
            self.navigation_bar.back_btn.setEnabled(False)

        elif self.possible_action == PossibleActions.Terminate:
            logger.info("Terminating current data collection")
            self.terminate.emit()
            self.possible_action = PossibleActions.CollectData
            self.calibrate_collect_btn.setText(self.possible_action.value)

        elif self.possible_action == PossibleActions.Calibrate:
            logger.info("Prematurely end data collection to initiate calibration")
            self.paired_frame_builder.store_points.clear()
            self.initiate_calibration()
            


    def enable_calibration(self):
        self.possible_action = PossibleActions.Calibrate
        self.calibrate_collect_btn.setText(self.possible_action.value)
        self.calibrate_collect_btn.setEnabled(True)
        
        
    def ImageUpdateSlot(self, q_image):
        self.stereo_frame_display.resize(self.stereo_frame_display.sizeHint())

        qpixmap = QPixmap.fromImage(q_image)
        self.stereo_frame_display.setPixmap(qpixmap)
        


    def initiate_calibration(self):
        def worker():
            self.calibration_initiated.emit()
            logger.info("Beginning wind-down process prior to calibration")
            self.calibrate_collect_btn.setText("---calibrating---")
            self.calibrate_collect_btn.setEnabled(False)
            # self.frame_emitter.stop()
            # self.stereo_frame_display.hide()
            logger.info("Stop recording video")
            self.session.stop_recording()
            logger.info("Begin calibration")
            logger.info("Pause synchronizer")
            self.session.pause_synchronizer()
            self.session.estimate_extrinsics()

            self.possible_action = PossibleActions.CollectData
            self.calibrate_collect_btn.setText(self.possible_action.value)
            self.calibrate_collect_btn.setEnabled(True)
            self.calibration_complete.emit()
            logger.info("Calibration Complete signal sent...")
            
        self.init_calibration_thread = Thread(target=worker,args=(), daemon=True)
        self.init_calibration_thread.start()

class PairedFrameEmitter(QThread):
    ImageBroadcast = pyqtSignal(QImage)
    calibration_data_collected = pyqtSignal() 
    possible_to_initialize_array = pyqtSignal()
    
    def __init__(self, paired_frame_builder:PairedFrameBuilder):
        
        super(PairedFrameEmitter,self).__init__()
        self.paired_frame_builder = paired_frame_builder
        logger.info("Initiated frame emitter")        
        self.keep_collecting = Event() 
        self.collection_complete = False
        
    def run(self):

        self.keep_collecting.set()
        self.collection_complete = False

        possible_to_initialize = False
        
        while self.keep_collecting.is_set():
            
            # that that it is important to make sure that this signal is sent only once
            # to avoid multiple calibration attempts 
            if len(self.paired_frame_builder.stereo_list) == 0 and not self.collection_complete:
                logger.info("Signalling that calibration data is fully collected.")
                self.collection_complete = True
                self.calibration_data_collected.emit()
        
                # break
            
            if not possible_to_initialize:
                # check to see if it is now
                if self.paired_frame_builder.possible_to_initialize_array(MIN_THRESHOLD_FOR_EARLY_CALIBRATE):
                    logger.info("Signaling that it is possible to initialize array based on collected data.")
                    possible_to_initialize = True
                    self.possible_to_initialize_array.emit()
                      
            stereo_frame = self.paired_frame_builder.get_stereo_frame()

            if stereo_frame is not None:
                image = cv2_to_qlabel(stereo_frame)
                self.ImageBroadcast.emit(image)

        logger.info("Stereoframe emitter run thread ended...") 
            
    # def stop(self):
        # self.keep_collecting.clear() 


        
        
        
def cv2_to_qlabel(frame):
    image = cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    qt_frame = QImage(
        image.data,
        image.shape[1],
        image.shape[0],
        QImage.Format.Format_RGB888,
    )
    return qt_frame

    
if __name__ == "__main__":
        from pyxy3d.configurator import Configurator
        from pyxy3d.trackers.charuco_tracker import CharucoTracker

        App = QApplication(sys.argv)

        session_path = Path(__root__, "dev","sample_sessions", "257")
        configurator = Configurator(session_path)

        session = Session(configurator)
        # session.load_cameras()
        tracker = CharucoTracker(session.charuco)
        session.load_stream_tools(tracker=tracker)


        stereo_dialog = ExtrinsicCalibrationWidget(session)
        stereo_dialog.show()

        sys.exit(App.exec())




import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)
# import logging
# logger.setLevel(logging.DEBUG)
import time
from queue import Queue
from threading import Thread, Event

import cv2
import numpy as np


import pyxy3d.calibration.draw_charuco as draw_charuco
from pyxy3d.calibration.charuco import Charuco
from pyxy3d.trackers.charuco_tracker import CharucoTracker
from pyxy3d.interface import FramePacket
from pyxy3d.cameras.live_stream import LiveStream

class MonoCalibrator():

    def __init__(
        self, stream:LiveStream,  board_threshold=0.7, wait_time=0.5, fps = 6
    ):
        self.stream = stream
        self.camera: Camera = stream.camera  # reference needed to update params
        self.port = self.camera.port
        self.board_threshold = board_threshold
                
        # # this is strange but is done to allow the GUI an easy way to restore stream fps when switching between monocal/synchronizer
        # # self.fps = fps # monocalibrator should not store fps...it's a stream prop. 
        # self.set_stream_fps(fps)

        self.wait_time = wait_time
        self.capture_corners = Event()
        self.capture_corners.clear() # start out not doing anything
        self.stop_event = Event()
                
        self.frame_packet_in_q = Queue(-1)    
        self.subscribe_to_stream()

        self.grid_frame_ready_q = Queue()


        self.initialize_grid_history()

        self.last_calibration_time = (
            time.perf_counter()
        )  # need to initialize to *something*

        self.thread = Thread(target=self.collect_corners, args=(), daemon=True)
        self.thread.start()

        logger.info(f"Beginning monocalibrator for port {self.port}")

    # def set_stream_fps(self, fps_target:int=None):
    #     """
    #     If new target, update monocal property, otherwise revert to previously stored
    #     """
    #     if fps_target is not None:
    #         self.fps = fps_target
    #     self.stream.set_fps_target(self.fps)
     
    @property
    def grid_count(self):
        """How many sets of corners have been collected up to this point"""
        return len(self.all_ids)

    @property
    def image_size(self):
        image_size = list(self.camera.size)
        image_size.reverse()  # for some reason...
        image_size.append(3)

        return image_size

    def initialize_grid_history(self):
        self.grid_capture_history = np.zeros(self.image_size, dtype="uint8")

        # roll back collected corners to the beginning
        self.all_ids = []
        self.all_img_loc = []
        self.all_obj_loc = []
    
    def stop(self):
        self.stop_event.set()
        self.thread.join()

    def subscribe_to_stream(self):
        self.stream.subscribe(self.frame_packet_in_q)
    
    def unsubscribe_to_stream(self): 
        self.stream.unsubscribe(self.frame_packet_in_q)
    

    def collect_corners(self):
        """
        Input: opencv frame

        Primary Action: records corner ids, positions, and board positions provided
        that enough time has past since the last set was recorded

        """
        logger.info(f"Entering collect_corners thread loop for port {self.port}")

        # these values are getting initialized now because the stream may not have
        # a tracker at the time the stream tools are loaded
        self.connected_points = self.stream.tracker.get_connected_points()
        board_corner_count = len(self.stream.tracker.board.getChessboardCorners())
        self.min_points_to_process = int(board_corner_count * self.board_threshold)
        
        while not self.stop_event.is_set():
            
            self.frame_packet: FramePacket = self.frame_packet_in_q.get()
            self.frame = self.frame_packet.frame

            if self.capture_corners.is_set() and self.frame_packet.points is not None:
                self.ids = self.frame_packet.points.point_id
                self.img_loc = self.frame_packet.points.img_loc
                self.obj_loc = self.frame_packet.points.obj_loc

                if self.ids.any():
                    enough_corners = len(self.ids) > self.min_points_to_process
                else:
                    enough_corners = False

                enough_time_from_last_cal = (
                    time.perf_counter() > self.last_calibration_time + self.wait_time
                )

                if enough_corners and enough_time_from_last_cal:
                    logger.debug(f"Points found and being processed for port {self.port}")

                    # store the corners and IDs
                    self.all_ids.append(self.ids)
                    self.all_img_loc.append(self.img_loc)
                    self.all_obj_loc.append(self.obj_loc)

                    self.last_calibration_time = time.perf_counter()
                    self.update_grid_history()
                else:
                    logger.debug(f"No points collected for processing at port {self.port}")
                    
                    
            if self.frame_packet.frame is not None:
                self.set_grid_frame()

        logger.info(f"Monocalibrator at port {self.port} successfully shutdown...")
        self.stream.push_to_out_q.clear()
        
    def update_grid_history(self):
        if len(self.ids) > 2:
            self.grid_capture_history = draw_charuco.grid_history(
                self.grid_capture_history,
                self.ids,
                self.img_loc,
                self.connected_points,
            )


    def set_grid_frame(self):
        """Merges the current frame with the currently detected corners (red circles) 
        and a history of the stored grid information."""

        logger.debug(f"Frame Size is {self.frame.shape} at port {self.port}")
        logger.debug(
            f"camera resolution is {self.camera.size} at port {self.port}"
        )

        # check to see if the camera resolution changed from the last round
        if (
            self.frame.shape[0] == self.grid_capture_history.shape[0]
            and self.frame.shape[1] == self.grid_capture_history.shape[1]
        ):
            self.frame_packet.frame = cv2.addWeighted(self.frame_packet.frame, 1, self.grid_capture_history, 1, 0)
            draw_charuco.corners(self.frame_packet)

            self.grid_frame = self.frame_packet.frame
            self.grid_frame_ready_q.put("frame ready")

        else:
            logger.debug("Reinitializing Grid Capture History")
            self.initialize_grid_history()
            self.grid_frame = self.grid_capture_history
            self.grid_frame_ready_q.put("frame ready")

    def calibrate(self):
        """
        Use the recorded image corner positions along with the objective
        corner positions based on the board definition to calculated
        the camera matrix and distortion parameters
        """
        logger.info(f"Calibrating camera {self.camera.port}....")

        objpoints = self.all_obj_loc
        imgpoints = self.all_img_loc
        height = self.grid_capture_history.shape[0]
        width = self.grid_capture_history.shape[1]

        self.error, self.mtx, self.dist, self.rvecs, self.tvecs = cv2.calibrateCamera(
            objpoints, imgpoints, (width, height), None, None
        )

        # fix extra dimension in return value of cv2.calibrateCamera
        self.dist = self.dist[0]

        self.update_camera()
        self.is_calibrated = True

        logger.info(f"Error: {self.error}")
        logger.info(f"Camera Matrix: {self.mtx}")
        logger.info(f"Distortion: {self.dist}")
        logger.info(f"Grid Count: {self.camera.grid_count}")

    def update_camera(self):
        logger.info(f"Setting calibration params on camera {self.camera.port}")
        # ret is RMSE of reprojection
        self.camera.error = round(self.error, 3)
        self.camera.matrix = self.mtx
        self.camera.distortions = self.dist
        self.camera.grid_count = len(self.all_ids)


if __name__ == "__main__":

    from pyxy3d.cameras.camera import Camera
    from pyxy3d.cameras.synchronizer import Synchronizer
    from pyxy3d.cameras.live_stream import LiveStream

    charuco = Charuco(
        3, 4, 11, 8.5, aruco_scale=0.75, square_size_overide_cm=5.25, inverted=True
    )
    charuco_tracker = CharucoTracker(charuco)


    test_port = 0
    cam = Camera(0)
    stream = LiveStream(cam, tracker=charuco_tracker)
    stream._show_fps = True
    # syncr = Synchronizer(streams, fps_target=20)

    monocal = MonoCalibrator(stream)

    monocal.capture_corners.set()
    
    print("About to enter main loop")
    while True:
        # read_success, frame = cam.capture.read()
        frame_ready = monocal.grid_frame_ready_q.get()
        logger.debug("Getting grid frame to display")
        frame = monocal.grid_frame

        cv2.imshow("Press 'q' to quit", frame)
        key = cv2.waitKey(1)

        # end capture when enough grids collected
        if key == ord("q"):
            cam.capture.release()
            cv2.destroyAllWindows()
            break

        if key == ord("t"):
            if monocal.stream.track_points.is_set():
                monocal.stream.track_points.clear()
            else:
                monocal.stream.track_points.set()
    
        if key == ord("v"):
            stream.change_resolution((1280,720))

    monocal.calibrate()
    monocal.update_camera()

    print(f"Error: {cam.error}")
    print(f"Camera Matrix: {cam.matrix}")
    print(f"Distortions: {cam.distortions}")
    print(f"Grid Count: {cam.grid_count}")



# this class is only a way to hold data related to the stereocamera triangulation.
# These will load from a config file (.toml) and provide a way for the 3D triangulation
# and plotting to manage the parameters. It feels like some duplication of the camera object,
# but I want something that is designed to be simple and not actually manage the cameras, just
# organize the saved data

import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

from queue import Queue
from threading import Thread, Event
from dataclasses import dataclass
import cv2
import numpy as np
import pandas as pd
from pathlib import Path
from itertools import combinations

from pyxy3d.triangulate.stereo_points_builder import StereoPointsBuilder, StereoPointsPacket
from pyxy3d.interface import PointPacket, FramePacket, SyncPacket
from pyxy3d.triangulate.stereo_points_builder import StereoPointsPacket, SynchedStereoPointsPacket


from pyxy3d.cameras.camera_array import CameraData, CameraArray


class ArrayStereoTriangulator:
    
    def __init__(self, camera_array: CameraArray):
        self.camera_array = camera_array
        
        # pull ports list from camera_array.port_index
        # to ensure only non-ignored cameras are processed
        self.ports = list(camera_array.port_index.keys())
        self.pairs = [(i,j) for i,j in combinations(self.ports,2) if i<j]


        # create the triangulators for each pair of cameras
        self.triangulators = {}
        for pair in self.pairs:
            port_A = pair[0]
            port_B = pair[1]
            
            camera_A:CameraData = self.camera_array.cameras[port_A]
            camera_B:CameraData = self.camera_array.cameras[port_B]

            self.triangulators[pair] = StereoPairTriangulator(camera_A, camera_B)
            
    def triangulate_synched_points(self, synced_paired_points:SynchedStereoPointsPacket):
        for pair,paired_point_packet  in synced_paired_points.stereo_points_packets.items():
            if paired_point_packet is not None:
                self.triangulators[pair].add_3D_points(paired_point_packet)



class StereoPairTriangulator:
    def __init__(self, camera_A: CameraData, camera_B: CameraData):

        self.camera_A = camera_A
        self.camera_B = camera_B
        self.portA = camera_A.port
        self.portB = camera_B.port
        self.pair = (self.portA, self.portB)

        self.build_projection_matrices()

    def build_projection_matrices(self):

        # attempting to create something that integrates with the new set_cameras_refactor
        # rot_A = np.linalg.inv(self.camera_A.rotation)
        # trans_A = np.array(self.camera_A.translation) * -1
        rot_trans_A = np.column_stack([self.camera_A.rotation, self.camera_A.translation])
        mtx_A = self.camera_A.matrix
        self.proj_A = mtx_A @ rot_trans_A  # projection matrix for CamA

        # rot_B = np.linalg.inv(self.camera_B.rotation)
        # trans_B = np.array(self.camera_B.translation) * -1
        rot_trans_B = np.column_stack([self.camera_B.rotation, self.camera_B.translation])
        mtx_B = self.camera_B.matrix
        self.proj_B = mtx_B @ rot_trans_B  # projection matrix for CamB

    def build_projection_matrices_old(self):

        # inversion/negation of R t here is legacy code that  
        # was based on my understanding at the time of frames of reference.
        # and it yields highly reasonable results. 
        # see https://stackoverflow.com/questions/17210424/3d-camera-coordinates-to-world-coordinates-change-of-basis
        # for a potential explanation. 
        # I would expect this to be a more common topic for computer vision forums
        # but I can't really find a reference to this and it bothers me
        rot_A = np.linalg.inv(self.camera_A.rotation)
        trans_A = np.array(self.camera_A.translation) * -1
        rot_trans_A = np.column_stack([rot_A, trans_A])
        mtx_A = self.camera_A.matrix
        self.proj_A = mtx_A @ rot_trans_A  # projection matrix for CamA

        rot_B = np.linalg.inv(self.camera_B.rotation)
        trans_B = np.array(self.camera_B.translation) * -1
        rot_trans_B = np.column_stack([rot_B, trans_B])
        mtx_B = self.camera_B.matrix
        self.proj_B = mtx_B @ rot_trans_B  # projection matrix for CamB

    def add_3D_points(self, paired_points:StereoPointsPacket):
            
        if len(paired_points.common_ids) > 0:
            xy_A = paired_points.img_loc_A
            xy_B = paired_points.img_loc_B
         
        if xy_A.shape[0] > 0:

            points_A_undistorted = self.undistort(xy_A,self.camera_A)
            points_B_undistorted = self.undistort(xy_B,self.camera_B)

            # triangulate points outputs data in 4D homogenous coordinate system
            # note that these are in a world frame of reference
            xyzw_h = cv2.triangulatePoints(
                self.proj_A, self.proj_B, points_A_undistorted, points_B_undistorted
            )

            xyz_h = xyzw_h.T[:, :3]
            w = xyzw_h[3, :]
            xyz = np.divide(xyz_h.T, w).T  # convert to euclidean coordinates
        else:
            xyz = np.array([])

        # update the paired point packet with the 3d positions
        paired_points.xyz = xyz
        
        
    def undistort(self, points, camera: CameraData, iter_num=3):
        # implementing a function described here: https://yangyushi.github.io/code/2020/03/04/opencv-undistort.html
        # supposedly a better implementation than OpenCV
        k1, k2, p1, p2, k3 = camera.distortions
        fx, fy = camera.matrix[0, 0], camera.matrix[1, 1]
        cx, cy = camera.matrix[:2, 2]
        
        # note I just made an edit to transpose these...I believe this is a consequence
        # of the recent switch to the PointPacket in the processing pipeline
        x, y = points.T[0], points.T[1]
        # x, y = float(point[0]), float(point[1])

        x = (x - cx) / fx
        x0 = x
        y = (y - cy) / fy
        y0 = y

        for _ in range(iter_num):
            r2 = x**2 + y**2
            k_inv = 1 / (1 + k1 * r2 + k2 * r2**2 + k3 * r2**3)
            delta_x = 2 * p1 * x * y + p2 * (r2 + 2 * x**2)
            delta_y = p1 * (r2 + 2 * y**2) + 2 * p2 * x * y
            x = (x0 - delta_x) * k_inv
            y = (y0 - delta_y) * k_inv
        return np.array((x * fx + cx, y * fy + cy))




if __name__ == "__main__":

    from pyxy3d.recording.recorded_stream import RecordedStreamPool
    from pyxy3d.cameras.synchronizer import Synchronizer
    from pyxy3d.calibration.charuco import Charuco
    from pyxy3d.trackers.charuco_tracker import CharucoTracker
    from pyxy3d.cameras.camera_array import CameraArrayBuilder, CameraArray, CameraData

    repo = Path(str(Path(__file__)).split("pyxy")[0],"pyxy")

    config_path = Path(repo, "sessions", "iterative_adjustment", "config.toml")
    camera_array = CameraArrayBuilder(config_path).get_camera_array()

    # create playback streams to provide to synchronizer
    recorded_data = Path(repo, "sessions", "iterative_adjustment", "recording")
    ports = [0, 1]
    recorded_stream_pool = RecordedStreamPool(ports, recorded_data)
    syncr = Synchronizer(
        recorded_stream_pool.streams, fps_target=None
    )  # no fps target b/c not playing back for visual display
    recorded_stream_pool.play_videos()

    # create a corner tracker to locate board corners
    tracker = Charuco(
        4, 5, 11, 8.5, aruco_scale=0.75, square_size_overide_cm=5.25, inverted=True
    )
    trackr = CharucoTracker(tracker)

    # create a commmon point finder to grab charuco corners shared between the pair of ports
    pairs = [(0, 1)]
    point_stream = StereoPointsBuilder(
        synchronizer=syncr,
        pairs=pairs,
        tracker=trackr,
    )

    camA, camB = camera_array.cameras[0], camera_array.cameras[2]
    pair = (camA.port, camB.port)

    # test_pair_out_q = Queue(-1)
    triangulatr = StereoPairTriangulator(camA, camB)
    frames_processed = 0

    while True:
        paired_points = point_stream.out_q.get()
        if paired_points.pair == (0, 1):
            triangulatr.in_q.put(paired_points)

        # print(all_pairs_common_points)
        # pair_points = all_pairs_common_points[pair]
        # if pair_points is not None:
        # triangulatr.in_q.put(paired_points)
        packet_3d = triangulatr.out_q.get()
        print(packet_3d.to_dict())
        frames_processed += 1
        # print(f"Frames Processed: {frames_processed}")
        # print(f"Time: {packet_3d.time}")



import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import sys
import math
from pathlib import Path
from threading import Thread, Event
import numpy as np
from queue import Queue
import pandas as pd
from pyxy3d.trackers.tracker_enum import TrackerEnum
from pyxy3d.export import xyz_to_trc

import cv2
from PyQt6.QtCore import Qt, pyqtSignal, QThread, QUrl
from PyQt6.QtGui import QDesktopServices, QImage, QPixmap, QIcon
from PyQt6.QtWidgets import (
    QApplication,
    QSizePolicy,
    QWidget,
    QProgressBar,
    QSpinBox,
    QScrollArea,
    QComboBox,
    QCheckBox,
    QTextEdit,
    QLineEdit,
    QListWidget,
    QGroupBox,
    QDoubleSpinBox,
    QHBoxLayout,
    QLabel,
    QPushButton,
    QSlider,
    QVBoxLayout,
)

from pyxy3d.post_processor import PostProcessor
from pyxy3d.session.session import Session
from pyxy3d.cameras.synchronizer import Synchronizer
from pyxy3d import __root__
from pyxy3d.recording.video_recorder import VideoRecorder
from pyxy3d.configurator import Configurator
from pyxy3d.gui.vizualize.playback_triangulation_widget import (
    PlaybackTriangulationWidget,
)
from pyxy3d.gui.progress_dialog import ProgressDialog

class PostProcessingWidget(QWidget):
    processing_complete = pyqtSignal()

    def __init__(self, session:Session):
        super(PostProcessingWidget, self).__init__()
        self.session = session
        self.config = session.config

        self.post_processor = PostProcessor(self.config)
        self.sync_index_cursors = {}
        self.recording_folders = QListWidget()

        self.update_recording_folders()

        # select the first element of the QListWidget
        if self.recording_folders.count() > 0:
            self.recording_folders.setCurrentRow(0)
            self.config = Configurator(self.active_recording_path)
            camera_array = self.config.get_camera_array()
            self.vis_widget = PlaybackTriangulationWidget(camera_array)
        else:
            raise RuntimeError("No recording folders, so cannot display anything in Post Processing Widget")
                

        self.tracker_combo = QComboBox()

        self.vizualizer_title = QLabel()

        # Add items to the combo box using the name attribute of the TrackerEnum
        for tracker in TrackerEnum:
            if tracker.name != "CHARUCO":
                self.tracker_combo.addItem(tracker.name, tracker)

        self.process_current_btn = QPushButton("&Process")
        # self.export_btn = QPushButton("&Export")
        self.open_folder_btn = QPushButton("&Open Folder")

        self.refresh_visualizer() # must happen before placement to create vis_widget and vizualizer_title
        self.place_widgets()
        self.connect_widgets()
        

    def set_current_xyz(self):

        if self.processed_xyz_path.exists():
            logger.info(f"Setting xyz display coordinates to those stored in {self.processed_xyz_path}")
            self.xyz = pd.read_csv(self.processed_xyz_path)
        else:
            logger.info(f"No points displayed; Nothing stored in {self.processed_xyz_path}")
            self.xyz = None
        self.vis_widget.set_xyz(self.xyz)

    def update_recording_folders(self):
        # this check here is an artifact of the way that the main widget handles refresh
        self.recording_folders.clear()

        # create list of recording directories
        dir_list = [p.stem for p in self.session.path.iterdir() if p.is_dir()]
        try:
            dir_list.remove("calibration")
        except:
            pass

        # add each folder to the QListWidget
        for folder in dir_list:
            self.recording_folders.addItem(folder)

    @property
    def processed_subfolder(self):
        subfolder = Path(
            self.session.path,
            self.recording_folders.currentItem().text(),
            self.tracker_combo.currentData().name,
        )
        return subfolder

    @property
    def processed_xyz_path(self):
        file_name = f"xyz_{self.tracker_combo.currentData().name}.csv"
        result = Path(self.processed_subfolder, file_name)
        return result

    def current_selection_processed(self) -> bool:
        """ "
        checks to see if their is a file in the recording directory named `xyz_TRACKERNAME.csv`
        """

        xyz_output = f"xyz_{self.tracker_combo.currentData().name}.csv"
        target_path = Path(self.processed_subfolder, xyz_output)

        return target_path.exists()

    @property
    def active_folder(self):
        if self.recording_folders.currentItem() is not None:
            active_folder: str = self.recording_folders.currentItem().text()
        else:
            active_folder = None
        return active_folder

    @property
    def active_recording_path(self)-> Path:
        p = Path(self.session.path, self.active_folder)
        logger.info(f"Active recording path is {p}")
        return p        
        
    @property
    def viz_title_html(self):
        if self.processed_xyz_path.exists():
            suffix = "(x,y,z) estimates"
        else:
            suffix = "(no processed data)"

        title = f"<div align='center'><b>{self.tracker_combo.currentData().name.title()} Tracker: {self.active_folder} {suffix} </b></div>"

        return title

    def place_widgets(self):
        self.setLayout(QHBoxLayout())
        self.left_vbox = QVBoxLayout()
        self.right_vbox = QVBoxLayout()
        self.button_hbox = QHBoxLayout()

        self.layout().addLayout(self.left_vbox)

        self.left_vbox.addWidget(self.recording_folders)
        self.left_vbox.addWidget(self.open_folder_btn)
        self.left_vbox.addWidget(self.tracker_combo)
        self.button_hbox.addWidget(self.process_current_btn)
        # self.button_hbox.addWidget(self.export_btn)
        self.left_vbox.addLayout(self.button_hbox)

        self.layout().addLayout(self.right_vbox, stretch=2)
        self.right_vbox.addWidget(self.vizualizer_title)
        self.right_vbox.addWidget(self.vis_widget, stretch=2)

    def connect_widgets(self):
        self.recording_folders.currentItemChanged.connect(self.refresh_visualizer)
        self.tracker_combo.currentIndexChanged.connect(self.refresh_visualizer)
        self.vis_widget.slider.valueChanged.connect(self.store_sync_index_cursor)
        self.process_current_btn.clicked.connect(self.process_current)
        self.open_folder_btn.clicked.connect(self.open_folder)
        # self.export_btn.clicked.connect(self.export_current_file)
        self.processing_complete.connect(self.enable_all_inputs)
        self.processing_complete.connect(self.refresh_visualizer)

    def store_sync_index_cursor(self, cursor_value):
        if self.processed_xyz_path.exists():
            self.sync_index_cursors[self.processed_xyz_path] = cursor_value
            # logger.info(self.sync_index_cursors)
        else:
            # don't bother, doesn't exist
            pass

    def open_folder(self):
        """Opens the currently active folder in a system file browser"""
        if self.active_folder is not None:
            folder_path = Path(self.session.path, self.active_folder)
            url = QUrl.fromLocalFile(str(folder_path))
            QDesktopServices.openUrl(url)
        else:
            logger.warn("No folder selected")

    def process_current(self):
        recording_path = Path(self.session.path, self.active_folder)
        logger.info(f"Beginning processing of recordings at {recording_path}")
        tracker_enum = self.tracker_combo.currentData()
        logger.info(f"(x,y) tracking will be applied using {tracker_enum.name}")
        recording_config_toml  = Path(recording_path,"config.toml")
        logger.info(f"Camera data based on config file saved to {recording_config_toml}")


        
        def processing_worker():
            logger.info(f"Beginning to process video files at {recording_path}")
            active_config = Configurator(recording_path) 
            logger.info(f"Creating post processor using config.toml in {recording_path}")
            self.post_processor = PostProcessor(active_config)

            self.disable_all_inputs()

            self.post_processor.create_xyz(recording_path,tracker_enum)
            trc_path = Path(
                self.processed_xyz_path.parent, self.processed_xyz_path.stem + ".trc"
            )
            logger.info(f"Saving data to {trc_path.parent}")

            # A side effect of the following line is that it also creates a wide labelled csv format
            xyz_to_trc(
                self.processed_xyz_path, self.tracker_combo.currentData().value()
            )
            self.processing_complete.emit()

        thread = Thread(target=processing_worker, args=(), daemon=True)
        thread.start()


    def refresh_visualizer(self):
        # logger.info(f"Item {item.text()} selected and double-clicked.")
        active_config = Configurator(self.active_recording_path)
        logger.info(f"Refreshing vizualizer to display camera array stored in config.toml in {self.active_recording_path}")
        camera_array = active_config.get_camera_array()
        
        self.vis_widget.update_camera_array(camera_array)
        
        self.set_current_xyz()
        self.vizualizer_title.setText(self.viz_title_html)
        self.update_enabled_disabled()
        self.update_slider_position()

    def disable_all_inputs(self):
        """used to toggle off all inputs will processing is going on"""
        self.recording_folders.setEnabled(False)
        self.tracker_combo.setEnabled(False)
        # self.export_btn.setEnabled(False)
        self.process_current_btn.setEnabled(False)
        self.vis_widget.slider.setEnabled(False)

    def enable_all_inputs(self):
        """
        after processing completes, swithes everything on again,
        but fine tuning of enable/disable will happen with self.update_enabled_disabled
        """
        self.recording_folders.setEnabled(True)
        self.tracker_combo.setEnabled(True)
        # self.export_btn.setEnabled(True)
        self.process_current_btn.setEnabled(True)
        self.vis_widget.slider.setEnabled(True)

    def update_enabled_disabled(self):
        if self.processed_xyz_path.exists():
            # self.export_btn.setEnabled(True)
            self.process_current_btn.setEnabled(False)
            self.vis_widget.slider.setEnabled(True)

        else:
            # self.export_btn.setEnabled(False)
            self.process_current_btn.setEnabled(True)
            self.vis_widget.slider.setEnabled(False)

    def update_slider_position(self):
        # update slider value to stored value if it exists
        if self.processed_xyz_path in self.sync_index_cursors.keys():
            active_sync_index = self.sync_index_cursors[self.processed_xyz_path]
            self.vis_widget.slider.setValue(active_sync_index)
            self.vis_widget.visualizer.display_points(active_sync_index)
        else:
            pass






import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)
import os
from numba import jit
from numba.typed import Dict, List
import numpy as np
import pandas as pd
from pyxy3d.cameras.camera_array import CameraArray
from pyxy3d.cameras.synchronizer import Synchronizer, SyncPacket
from queue import Queue
from threading import Thread, Event
from pathlib import Path
from pyxy3d.interface import XYZPacket, Tracker


class SyncPacketTriangulator:
    """
    Will place 3d packets on subscribed queues and save consolidated data in csv
    format to output_path if provided
    """

    def __init__(
        self,
        camera_array: CameraArray,
        synchronizer: Synchronizer,
        recording_directory: Path = None,
        tracker_name:str = None,  # used only for getting the point names and tracker name
    ):
        self.camera_array = camera_array
        self.synchronizer = synchronizer
        self.recording_directory = recording_directory

        self.stop_thread = Event()
        self.stop_thread.clear()

        self.tracker_name = tracker_name

        self.xyz_history = {
            "sync_index": [],
            "point_id": [],
            "x_coord": [],
            "y_coord": [],
            "z_coord": [],
        }

        self.sync_packet_in_q = Queue(-1)
        self.synchronizer.subscribe_to_sync_packets(self.sync_packet_in_q)

        # assemble numba compatible dictionary
        self.projection_matrices = Dict()
        # self.projection_matrices = {}
        for port, cam in self.camera_array.cameras.items():
            self.projection_matrices[port] = cam.projection_matrix

        self.subscribers = []
        self.running = True
        self.thread = Thread(target=self.process_incoming, args=(), daemon=True)
        self.thread.start()

    def subscribe(self, queue: Queue):
        self.subscribers.append(queue)

    def unsubscriber(self, queue: Queue):
        self.subscribers.remove(queue)

    def process_incoming(self):
        # waiting to set running property here was causing issues with identifying state of thread.
        # set property to true then start thread...

        # self.running = True
        while not self.stop_thread.is_set():
            sync_packet: SyncPacket = self.sync_packet_in_q.get()

            if sync_packet is None:
                # No more sync packets after this... wind down
                self.stop_thread.set()
                logger.info(
                    "End processing of incoming sync packets...end signaled with `None` packet"
                )
            else:
                logger.debug(
                    f"Sync Packet {sync_packet.sync_index} acquired with {sync_packet.frame_packet_count} frames"
                )
                # only attempt to process if data exists
                if sync_packet.frame_packet_count >= 2:
                    cameras, point_ids, imgs_xy = sync_packet.triangulation_inputs
                    logger.debug("Attempting to triangulate synced frames")
                    # prepare for jit
                    cameras = np.array(cameras)
                    point_ids = np.array(point_ids)
                    imgs_xy = np.array(imgs_xy)

                    logger.debug(f"Cameras are {cameras} and point_ids are {point_ids}")
                    if len(np.unique(cameras)) >= 2:
                        logger.debug(f"Points observed on cameras {np.unique(cameras)}")
                        point_id_xyz, points_xyz = triangulate_sync_index(
                            self.projection_matrices, cameras, point_ids, imgs_xy
                        )

                        logger.debug(
                            f"Synch Packet {sync_packet.sync_index} | Point ID: {point_id_xyz} | xyz: {points_xyz}"
                        )

                        xyz_packet = XYZPacket(
                            sync_packet.sync_index, point_id_xyz, points_xyz
                        )
                        logger.info(
                            f"Placing xyz pacKet for index {sync_packet.sync_index} with {len(xyz_packet.point_ids)} points"
                        )
                        for q in self.subscribers:
                            q.put(xyz_packet)

                        # if self.output_path is not None:
                        self.add_packet_to_history(xyz_packet)

        self.running = False

        if self.recording_directory is not None:
            logger.info(f"Saving xyz point data to {self.recording_directory}")
            self.save_history()

    def add_packet_to_history(self, xyz_packet: XYZPacket):
        point_count = len(xyz_packet.point_ids)

        if point_count > 0:
            self.xyz_history["sync_index"].extend([xyz_packet.sync_index] * point_count)
            xyz_array = np.array(xyz_packet.point_xyz)
            self.xyz_history["point_id"].extend(xyz_packet.point_ids)
            self.xyz_history["x_coord"].extend(xyz_array[:, 0].tolist())
            self.xyz_history["y_coord"].extend(xyz_array[:, 1].tolist())
            self.xyz_history["z_coord"].extend(xyz_array[:, 2].tolist())
            
    

    def save_history(self)->None:
        """
        If a recording directory is provided, then save the xyz directory into it
        If a tracker name is provided, then base name on the tracker name
        """
        df_xyz: pd.DataFrame = pd.DataFrame(self.xyz_history)
       
        if self.recording_directory is not None: 
            if self.tracker_name is None:
                filename = "xyz.csv"
            else:
                filename = f"xyz_{self.tracker_name}.csv"
                df_xyz.to_csv(Path(self.recording_directory,filename))


# helper function to avoid use of np.unique(return_counts=True) which doesn't work with jit
@jit(nopython=True, cache=True)
def unique_with_counts(arr):
    sorted_arr = np.sort(arr)
    unique_values = [sorted_arr[0]]
    counts = [1]

    for i in range(1, len(sorted_arr)):
        if sorted_arr[i] != sorted_arr[i - 1]:
            unique_values.append(sorted_arr[i])
            counts.append(1)
        else:
            counts[-1] += 1

    return np.array(unique_values), np.array(counts)


@jit(nopython=True, parallel=True, cache=True)
def triangulate_sync_index(
    projection_matrices, current_camera_indices, current_point_id, current_img
):
    # sync_indices_xyz = List()
    point_indices_xyz = List()
    obj_xyz = List()

    unique_points, point_counts = unique_with_counts(current_point_id)
    for index in range(len(point_counts)):
        if point_counts[index] > 1:
            # triangulate that point...
            point = unique_points[index]
            points_xy = current_img[current_point_id == point]
            camera_ids = current_camera_indices[current_point_id == point]

            num_cams = len(camera_ids)
            A = np.zeros((num_cams * 2, 4))
            for i in range(num_cams):
                x, y = points_xy[i]
                P = projection_matrices[camera_ids[i]]
                A[(i * 2) : (i * 2 + 1)] = x * P[2] - P[0]
                A[(i * 2 + 1) : (i * 2 + 2)] = y * P[2] - P[1]
            u, s, vh = np.linalg.svd(A, full_matrices=True)
            point_xyzw = vh[-1]
            point_xyz = point_xyzw[:3] / point_xyzw[3]

            point_indices_xyz.append(point)
            obj_xyz.append(point_xyz)

    return point_indices_xyz, obj_xyz



import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

import math
import sys
import time
from pathlib import Path
from threading import Thread
from queue import Queue
import pandas as pd
import numpy as np
import pyqtgraph.opengl as gl

from random import random

from pyxy3d.gui.vizualize.camera_mesh import CameraMesh, mesh_from_camera
from pyxy3d.cameras.camera_array import CameraArray
from pyxy3d.calibration.capture_volume.capture_volume import CaptureVolume


class CaptureVolumeVisualizer:
    """
    Can except either a single camera array or a capture volume that includes
    point_estimates. If a capture volume is supplied, point positions can
    be played back.
    """

    def __init__(
        self, capture_volume: CaptureVolume = None, camera_array: CameraArray = None
    ):

        if camera_array is not None and capture_volume is None:
            self.camera_array = camera_array
            self.point_estimates = None
        else:
            self.capture_volume = capture_volume
            self.camera_array = capture_volume.camera_array
            self.point_estimates = self.capture_volume.point_estimates

        # constuct a scene
        self.scene = gl.GLViewWidget()
        self.scene.setCameraPosition(distance=4)  # the scene camera, not a real Camera
        self.sync_index = None

        self.refresh_scene()

    def refresh_scene(self):
        logger.info("refreshing capture volume scene")
        self.scene.clear()

        axis = gl.GLAxisItem()
        self.scene.addItem(axis)

        # build meshes for all cameras
        self.meshes = {}
        for port, cam in self.camera_array.cameras.items():
            print(port)
            print(cam)
            mesh:CameraMesh = mesh_from_camera(cam)
            self.meshes[port] = mesh
            self.scene.addItem(mesh)

        # self.scene.show()

        if self.point_estimates is not None:
            self.scatter = gl.GLScatterPlotItem(
                pos=np.array([0, 0, 0]),
                color=[1, 1, 1, 1],
                size=0.01,
                pxMode=False,
            )
            self.scene.addItem(self.scatter)

            self.sync_indices = np.unique(self.point_estimates.sync_indices)
            self.sync_indices = np.sort(self.sync_indices)

            self.min_sync_index = np.min(self.sync_indices)
            self.max_sync_index = np.max(self.sync_indices)
   
            if self.sync_index is not None:
                self.display_points(self.sync_index)
                 
    def display_points(self, sync_index):
        """
        sync_index is provided from the dialog and linked to the slider
        it is initially set to the minimum viable sync index
        """
        self.sync_index = sync_index
        current_sync_index_flag = self.point_estimates.sync_indices == sync_index
        single_board_indices = np.unique(
            self.point_estimates.obj_indices[current_sync_index_flag]
        )


        self.single_board_points = self.point_estimates.obj[single_board_indices]
        self.mean_board_position = np.mean(self.single_board_points,axis=0)
        logger.debug(f"Mean Board Position at sync index {sync_index}: {self.mean_board_position}")

        self.scatter.setData(pos=self.single_board_points)




# %%



import pyxy3d.logger
logger = pyxy3d.logger.get(__name__)

import sys
from pathlib import Path

from PyQt6.QtCore import pyqtSignal, Qt
from PyQt6.QtWidgets import (
    QApplication,
    QVBoxLayout,
    QWidget,
    QTabWidget,
)

from pyxy3d.gui.camera_config.camera_config_dialogue import CameraConfigTab
from pyxy3d.session.session import Session
from pyxy3d.gui.navigation_bars import NavigationBarBackNext

class IntrinsicCalibrationWidget(QWidget):
    """ 
    This is basically just the camera tabs plus the navigation bar
    """
    def __init__(self, session: Session):
        super(IntrinsicCalibrationWidget, self).__init__()
        self.setLayout(QVBoxLayout())    
        self.camera_tabs = CameraTabs(session)
        self.layout().addWidget(self.camera_tabs)
    
        self.session = session
            
class CameraTabs(QTabWidget):
    
    stereoframe_ready = pyqtSignal(bool)

    def __init__(self, session: Session):
        super(CameraTabs, self).__init__()
        self.session = session

        self.setTabPosition(QTabWidget.TabPosition.North)
        self.add_cam_tabs()
        # self.session.set_mode(SessionMode.IntrinsicCalibration)
        self.currentChanged.connect(self.activate_current_tab)
        self.session.active_monocalibrator = self.currentWidget().stream.port
        self.session.activate_monocalibrator()

    def keyPressEvent(self, event):
        """
        Override the keyPressEvent method to allow navigation via PgUp/PgDown
        """

        if event.key() == Qt.Key.Key_PageUp:
            current_index = self.currentIndex()
            if current_index > 0:
                self.setCurrentIndex(current_index - 1)
        elif event.key() == Qt.Key.Key_PageDown:
            current_index = self.currentIndex()
            if current_index < self.count() - 1:
                self.setCurrentIndex(current_index + 1)
        else:
            super().keyPressEvent(event)
        
        
    def activate_current_tab(self, index):

        logger.info(f"Toggle tracking to activate {self.tabText(index)}")
        self.session.pause_all_monocalibrators()
        self.session.activate_monocalibrator(self.widget(index).stream.port)

        # this is where you can update the spin boxes to align with the session values
        monocal_fps = self.session.get_active_mode_fps()
        self.widget(index).advanced_controls.frame_rate_spin.setValue(monocal_fps)

        wait_time_intrinsic = self.session.wait_time_intrinsic
        self.widget(index).advanced_controls.wait_time_spin.setValue(wait_time_intrinsic)

    def add_cam_tabs(self):
        tab_names = [self.tabText(i) for i in range(self.count())]
        logger.info(f"Current tabs are: {tab_names}")

        if len(self.session.monocalibrators) > 0:
            
            # construct a dict of tabs so that they can then be placed in order
            tab_widgets = {}
            for port, monocal in self.session.monocalibrators.items():
                tab_name = f"Camera {port}"
                logger.info(f"Potentially adding {tab_name}")
                if tab_name in tab_names:
                    pass  # already here, don't bother
                else:
                    cam_tab = CameraConfigTab(self.session, port)
                    tab_widgets[port] = cam_tab

            # add the widgets to the tab bar in order
            ordered_ports = list(tab_widgets.keys())
            ordered_ports.sort()
            for port in ordered_ports:
                self.insertTab(port, tab_widgets[port], f"Camera {port}")
            
        else:
            logger.info("No cameras available")
        


import typing
import pyxy3d.logger

logger = pyxy3d.logger.get(__name__)

from time import sleep, time
from queue import Queue
import cv2
from PyQt6.QtCore import QObject, pyqtSignal

import sys
from PyQt6.QtWidgets import QApplication
from pyxy3d.configurator import Configurator
from pathlib import Path
import numpy as np
from numba.typed import Dict, List
from pyxy3d import __root__
import pandas as pd
from pyxy3d.cameras.camera_array import CameraArray
from pyxy3d.recording.recorded_stream import RecordedStream, RecordedStreamPool
from pyxy3d.cameras.synchronizer import Synchronizer
from pyxy3d.recording.video_recorder import VideoRecorder
from pyxy3d.triangulate.sync_packet_triangulator import (
    SyncPacketTriangulator,
    triangulate_sync_index,
)
from pyxy3d.interface import FramePacket, Tracker
from pyxy3d.trackers.tracker_enum import TrackerEnum
from pyxy3d.gui.playback_widget import PlaybackWidget

# specify a source directory (with recordings)
from pyxy3d.helper import copy_contents


class PostProcessor(QObject):
    progress_update = pyqtSignal(dict)  # {"stage": str, "percent":int}

    def __init__(self,config:Configurator ):
        super().__init__()
        self.config = config

    def create_xyz(self, recording_path: Path, tracker_enum: TrackerEnum,) -> None:
        """
        creates xyz_{tracker name}.csv file within the recording_path directory

        Uses the two functions above, first creating the xy points based on the tracker if they 
        don't already exist, the triangulating them. Makes use of an internal method self.triangulate_xy_data
        
        """

        output_suffix = tracker_enum.name

        tracker_output_path = Path(recording_path, tracker_enum.name)
        # locate xy_{tracker name}.csv
        xy_csv_path = Path(tracker_output_path, f"xy_{output_suffix}.csv")

        # create if it doesn't already exist
        if not xy_csv_path.exists():
            self.create_xy(recording_path, tracker_enum)

        # load in 2d data and triangulate it
        xy_data = pd.read_csv(xy_csv_path)
        xyz_history = self.triangulate_xy_data(xy_data)
        xyz_data = pd.DataFrame(xyz_history)
        xyz_data.to_csv(Path(tracker_output_path, f"xyz_{output_suffix}.csv"))

    def create_xy(self, recording_path: Path, tracker_enum: TrackerEnum,):
        """
        Reads through all .mp4  files in the recording path and applies the tracker to them
        The xy_TrackerName.csv file is saved out to the same directory by the VideoRecorder
        """
        frame_times = pd.read_csv(Path(recording_path, "frame_time_history.csv"))
        sync_index_count = len(frame_times["sync_index"].unique())


        logger.info("Creating pool of playback streams to begin processing")
        stream_pool = RecordedStreamPool(
            directory=recording_path,
            config=self.config,
            fps_target=100,
            tracker=tracker_enum.value(),
        )

        synchronizer = Synchronizer(stream_pool.streams, fps_target=100)

        logger.info(
            "Creating video recorder to record (x,y) data estimates from PointPacket delivered by Tracker"
        )
        output_suffix = tracker_enum.name
        video_recorder = VideoRecorder(synchronizer, suffix=output_suffix)

        # store video files in a subfolder named by the tracker_enum.name
        destination_folder = Path(recording_path, tracker_enum.name)
        video_recorder.start_recording(
            destination_folder=destination_folder,
            include_video=True,
            show_points=True,
            store_point_history=True
        )
        logger.info("Initiate playback and processing")
        stream_pool.play_videos()

        while video_recorder.recording:
            sleep(1)
            percent_complete = int((video_recorder.sync_index / sync_index_count) * 100)
            logger.info(f"(Stage 1 of 2): {percent_complete}% of frames processed for (x,y) landmark detection")
            self.progress_update.emit(
                {
                    "stage": "Estimating (x,y) landmark positions (stage 1 of 2)",
                    "percent": percent_complete,
                }
            )


    def triangulate_xy_data(self, xy_data: pd.DataFrame) -> Dict[str, List]:
        
        camera_array = self.config.get_camera_array()
        # assemble numba compatible dictionary
        projection_matrices = Dict()
        for port, cam in camera_array.cameras.items():
            projection_matrices[int(port)] = cam.projection_matrix

        xyz_history = {
            "sync_index": [],
            "point_id": [],
            "x_coord": [],
            "y_coord": [],
            "z_coord": [],
        }

        sync_index_max = xy_data["sync_index"].max()

        start = time()
        last_log_update = int(start)  # only report progress each second

        for index in xy_data["sync_index"].unique():
            active_index = xy_data["sync_index"] == index
            cameras = xy_data["port"][active_index].to_numpy()
            point_ids = xy_data["point_id"][active_index].to_numpy()
            img_loc_x = xy_data["img_loc_x"][active_index].to_numpy()
            img_loc_y = xy_data["img_loc_y"][active_index].to_numpy()
            imgs_xy = np.vstack([img_loc_x, img_loc_y]).T

            # the fancy part
            point_id_xyz, points_xyz = triangulate_sync_index(
                projection_matrices, cameras, point_ids, imgs_xy
            )

            if len(point_id_xyz) > 0:
                # there are points to store so store them...
                xyz_history["sync_index"].extend([index] * len(point_id_xyz))
                xyz_history["point_id"].extend(point_id_xyz)

                points_xyz = np.array(points_xyz)
                xyz_history["x_coord"].extend(points_xyz[:, 0].tolist())
                xyz_history["y_coord"].extend(points_xyz[:, 1].tolist())
                xyz_history["z_coord"].extend(points_xyz[:, 2].tolist())

            # only log percent complete each second
            if int(time()) - last_log_update >= 1:
                percent_complete = int(100*(index/sync_index_max))
                logger.info(
                    f"(Stage 2 of 2): Triangulation of (x,y) point estimates is {percent_complete}% complete"
                )
                last_log_update = int(time())
                self.progress_update.emit(
                    {
                        "stage": "Triangulating (x,y,z) estimates (stage 2 of 2)",
                        "percent": percent_complete,
                    }
                )

        # signalling the progress bar can now close
        self.progress_update.emit(
            {
                "stage": "Triangulating (x,y,z) estimates (stage 2 of 2)",
                "percent": 100,
                "close": True
            }
        )

        return xyz_history

